"""
Class to save/restore configuration from file
"""

# CM0004

import ast
import base64
import logging
import pickle
import xml
import xml.etree.ElementTree as ET
import xml.dom.minidom as DOM

from pathlib import Path

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class ConfigurationSettings:
    """
    Maintains a dictionary that can be saved to an
    xml file. It can restore the values reading the
    file.

    Examples:

    .. code-block:: python

        >>> cfg = ConfigurationSettings('file.xml')
        >>> value = "Configuration Value"
        >>> value
        'Configuration Value'
        >>> cfg.set('key', value)
        >>> cfg.get('key')
        'Configuration Value'
        >>> cfg.saveToFile()
        >>> cfg2 = ConfigurationSettings('file.xml')
        >>> cfg2.readFromFile()
        >>> value2 = cfg2.get('key')
        >>> value2
        'Configuration Value'

    Next example using a class

    .. code-block:: python

        >>> from vsutillib.files import ConfigurationSettings
        >>> class Abc():
        ...     def __init__(self, param):
        ...         self.value = param
        ...     def getValue(self):
        ...         return self.value
        ...
        >>> cfg = ConfigurationSettings('file.xml')
        >>> c = Abc(13)
        >>> c.getValue()
        13
        >>> cfg.set('class', c, valueType='pickle')
        >>> cfg.saveToFile()
        >>> cfg2 = ConfigurationSettings('file.xml')
        >>> cfg2.readFromFile()
        >>> c2 = cfg2.get('class')
        >>> c2.getValue()
        13

    keys have to be strings

    It works with basic data types:

        basic data types and lists, dictionaries,
        tuples or sets of these data types

        - bool, bytes, numbers: [int, float, complex], strings, set

        theese types can be saved but not in lists, dictionaries,
        tuples or sets

        - range, bytearray, frozenset, function


    Binary data or lists, dictionaries, tuple and sets with
    binary data can be tried with **valueType='pickle'**.  As seen
    in the second example using a class is possible but it does not
    always work.  The scope of the class ConfigurationSettings is
    for saving simple items that serve as a configuration setting
    for an application in different operating systems.  Is recommended
    not to use values that have to be pickled this will make it system
    dependent also it may be dependent of the Python version. That
    said binary values have to be througly tested.

    Args:
        configFile (str): xml file name or Path object

    Raises:
        ValueError: Raised when the xml file can not be created
                    when first run and the file still don't exist
    """

    # log state
    __log = False

    # data that can be treated as literal
    # theese are human readable easier
    # to use on different systems
    _literal = [
        "bool",
        "bytes",
        "complex",
        "float",
        "int",
        "str",
        "dict",
        "list",
        "tuple",
        "set",
    ]

    # pickable types are python specific maybe
    # even version specific
    _pickable = ["range", "bytearray", "frozenset", "function", "pickle"]

    @classmethod
    def classLog(cls, setLogging=None):
        """
        get/set logging at class level
        every class instance will log
        unless overwritten

        Args:
            setLogging (bool):
                - True class will log
                - False turn off logging
                - None returns current Value

        Returns:
            bool:

            returns the current value set
        """

        if setLogging is not None:
            if isinstance(setLogging, bool):
                cls.__log = setLogging

        return cls.__log

    def __init__(self, configFile=None):

        self._config = {}
        self._configType = {}
        self._configFile = configFile

        # for iteration
        self._current = 0
        self._len = 0

        # global logging override
        self.__log = None

    def __contains__(self, value):
        return value in self._config

    def __getitem__(self, key):
        return self._config[key]

    def __setitem__(self, key, value):
        self._config[key] = value

    def __iter__(self):
        return self

    def __next__(self):
        if self._current >= self._len:
            self._current = 0
            raise StopIteration
        else:
            self._current += 1
            key = list(self._config)[self._current - 1]
            return [key, self._config[key]]

    def __bool__(self):
        return len(self._config)

    def __len__(self):
        return len(self._config)

    @property
    def log(self):
        """
        class property can be used to override the class global
        logging setting

        Returns:
            bool:

            True if logging is enable

            False otherwise
        """
        if self.__log is not None:
            return self.__log

        return ConfigurationSettings.classLog()

    @property
    def configDictionary(self):
        return self._config

    @log.setter
    def log(self, value):
        """set instance log variable"""
        if isinstance(value, bool) or value is None:
            self.__log = value

    def set(self, key, value, valueType=None):
        """
        set value at key in dictionary

        Args:
            key (str): configuration element

            value (obj): element value the type is as explained

            valueType (str): specify the type to use to save value
        """

        if isinstance(key, str):
            self._config[key] = value
            _valueType = type(value).__name__

            if valueType is not None:
                _valueType = valueType

            if not ((_valueType in self._pickable) or (_valueType in self._literal)):
                s = str(_valueType)
                if self.log:
                    MODULELOG.debug("CFG0003: value type not supported - %s", str(s))
                raise TypeError("value type not supported - {}".format(s))

            self._configType[key] = _valueType

            self._len = len(self._config)

        else:
            s = str(key)
            if self.log:
                MODULELOG.debug("CFG0002: key must be a string - %s", str(s))
            raise TypeError("key must be a string - {}".format(s))

    def get(self, key):
        """
        get value from dictionary

        Args:
            key (str): configuration element

        Returns:
            object as explained
        """

        if key in self._config:
            return self._config[key]

        if self.log:
            s = str(key)
            MODULELOG.debug("CFG0001: key not found - %s", s)

        return None

    def delete(self, key):
        """
        delete remove value from dictionary

        Args:
            key (str): configuration element

        Returns:
            object removed None if key not found
        """

        retVal = self._config.pop(key, None)

        if retVal is None:
            if self.log:
                s = str(key)
                MODULELOG.debug("CFG0004: key not found - %s", s)

        return retVal

    def toXML(self, root=None, name=None):
        """
        Returns the configuration in XML format
        if root is None returns the current configuration

        Args:
            root (xml.etree.ElementTree): root of document

            name (str): root tag

        Returns:
            xml.etree.ElementTree
        """

        if name is None:
            name = "Config"

        config = ET.Element(name)

        if root is not None:
            root.append(config)

        for key, value in self:
            valueType = type(value).__name__

            if key in self._configType:
                valueType = self._configType[key]

            if valueType in self._literal:
                tValue = value
            elif valueType in self._pickable:
                p = pickle.dumps(value)
                u = base64.b64encode(p)
                tValue = u

            configElement = ET.SubElement(config, "ConfigSetting")
            configElement.attrib = {"id": key, "type": valueType}
            configElement.text = str(tValue)

        if root is None:
            return config

        return root

    def fromXML(self, xmlDoc, name=None):
        """
        Restore configuration from xml name parameter
        permit various configuration sets on same
        xml document

        Args:
            xmlDoc (xml.etree.ElementTree): xml document containing
            configuration data
        """
        self._config = {}

        if name is None:
            searchIn = "Config/ConfigSetting"
        else:
            searchIn = name + "/ConfigSetting"

        for setting in xmlDoc.findall(searchIn):

            key = setting.attrib["id"]
            valueType = setting.attrib["type"]

            if valueType == "str":
                value = setting.text
            elif valueType in self._pickable:
                u = ast.literal_eval(setting.text)
                value = pickle.loads(base64.b64decode(u))
            else:
                value = ast.literal_eval(setting.text)

            self.set(key, value, valueType=valueType)

    def xmlPrettyPrint(self, root=None):
        """
        Returns configuration xml Pretty Printed

        Args:
            root (xml.etree.ElementTree)

        Returns:
            xml.dom.minidom
        """

        if root is not None:
            if not isinstance(root, xml.etree.ElementTree.Element):
                return None
        else:
            root = self.toXML()

        xmlDoc = DOM.parseString(ET.tostring(root))

        xmlPretty = xmlDoc.toprettyxml(indent="    ")

        return xmlPretty

    def setConfigFile(self, xmlFile):
        """
        sets the file for reading and writing to
        """

        p = Path(xmlFile)

        if not p.anchor:
            xf = Path(Path.home(), xmlFile)
        else:
            xf = p

        self._configFile = xf

    def saveToFile(self, xmlFile=None, rootName=None):
        """
        save configuration to file in xml format

        Args:
            xmlFile (Path|str): file to write to

            rootName (str): root tag
        """

        xf = xmlFile
        if xmlFile is None:
            xf = self._configFile

        if rootName is None:
            rootTag = "VergaraSoft"

        root = ET.Element(rootTag)

        xmlConfig = self.toXML(root)
        tree = ET.ElementTree(xmlConfig)
        tree.write(str(xf))

    def readFromFile(self, xmlFile=None):
        """
        save configuration to file in xml format

        Args:
            xmlFile (Path|str): file to read

            rootName (str): root tag

        Raises:
            ValueError: raised when file con not be read
        """

        xf = xmlFile
        if xmlFile is None:
            xf = self._configFile

        f = Path(xf)

        if f.is_file():
            tree = ET.ElementTree(file=str(xf))
            root = tree.getroot()

            self.fromXML(root)


class Abc:
    """Test class"""

    def __init__(self, param):
        self.value = param

    def getValue(self):
        """Test method"""
        return self.value


def print13():
    """function for testing"""
    print("Print from function = 13")


def test():
    """Testing read and write configuration to file"""

    classInstance = Abc(13)

    configFile = Path(Path.cwd(), "configmanager.xml")

    configuration = ConfigurationSettings(configFile=configFile)

    b = b"Sanson"

    configuration.set("range", range(13))
    configuration.set("set", {"r", "n", "a", "f", "e", "i"})
    configuration.set("bytearray", bytearray(b"Itsue"))
    configuration.set("frozenset", frozenset("Itsue"))
    configuration.set("function", print13, valueType="pickle")
    configuration.set("class", classInstance, valueType="pickle")
    configuration.set("bool", True)
    configuration.set(
        "base64sting",
        "AdnQywACAAAAAAHmAAAAoAAACM4AAAR5AAAB7wAAAMYAAAjFAAAEcAAAAAAAAAAACgA=",
    )
    configuration.set(
        "base86bytes",
        "AdnQywACAAAAAAHmAAAAoAAACM4AAAR5AAAB7wAAAMYAAAjFAAAEcAAAAAAAAAAACgA=".encode(),
    )
    configuration.set("dict", {"key1": 1, "key2": 2, 3: b})
    configuration.set(
        "list", [2, 3, "list", {"key1": 1, 2: [2]}], valueType="pickle"
    )
    configuration.set("int", 13)
    configuration.set("float", 1.3e200)
    configuration.set("complex", 1 + 3j)
    configuration.set("tuple", (1.11, 2.22, 3.33))

    print("\nConfiguration set\n")
    for key, value in configuration:
        print(
            "Key = {0}, type = {2} value = {1}".format(key, value, type(value).__name__)
        )

    configuration.saveToFile()

    configuration.readFromFile()

    root = configuration.toXML()

    print("\nRead from configuration file\n")
    for key, value in configuration:
        print(
            "Key = {0}, type = {2}, value = {1}".format(
                key, value, type(value).__name__
            )
        )

    prettyXML = configuration.xmlPrettyPrint(root)

    print()
    print(prettyXML)
    print("Call function: ")
    f = configuration.get("function")
    f()
    c = configuration.get("class")
    print("Calling class method = {} ".format(c.getValue()))


if __name__ == "__main__":
    test()



""" QLogHandler """

import logging

from PySide2.QtCore import QObject, Signal


class Communicate(QObject):
    logRecord = Signal(object)


class QSignalLogHandler(logging.Handler):

    """
    logging handler that rotate files at initialization time
    this will create a new file each time is run against
    the same log file.

    Args:
        **kwargs: variable number of key value parameters
            that are passed to the super class on
            initialization
    """

    def __init__(self, slotFunction=None, **kwargs):
        super().__init__(**kwargs)

        self.signal = Communicate()

        if slotFunction is not None:
            self.connect(slotFunction)

    def emit(self, record):
        msg = self.format(record)
        self.signal.logRecord.emit(msg)

    def connect(self, slotFunction):

        self.signal.logRecord.connect(slotFunction)



"""
verify mkvmerge-gui for consistency
no errors should be found ever
except for wrong copy/paste
manual entry or edits after paste

MKVCommand will find the real errors
  number of source files not matching
"""
# VMC0001

import ast
import logging
import re

from pathlib import Path

from ..mkvutils import stripEncaseQuotes

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class VerifyMKVCommand:
    """
    Convenience class use by MKVCommand_

    Sanity check on command any failure results in no action whatsoever
    The class evaluates to True if no problems are found False otherwise.

    .. code:: Python

        verify = VerifyMKVCommand(command)

        if verify:
            # Ok to proceed
            ...
        else:
            raise ValueError('')


    Args:
        strCommand (:obj:`str`, optional): command line as generated by mkvtoolnix-gui
    """

    __log = False

    @classmethod
    def classLog(cls, setLogging=None):
        """
        get/set logging at class level
        every class instance will log
        unless overwritten

        Args:
            setLogging (`bool`):

                - True class will log
                - False turn off logging
                - None returns current Value

        Returns:
            bool:

            returns the current value set
        """

        if setLogging is not None:
            if isinstance(setLogging, bool):
                cls.__log = setLogging

        return cls.__log

    def __init__(self, strCommand=None, log=None):

        self.__lstAnalysis = None
        self.__errorFound = False
        self.__strCommand = None
        self.__bashCommand = None
        self.__outputFile = None
        self.__chaptersFile = None
        self.__log = None

        self.log = log

        if strCommand is not None:
            self.__strCommand = strCommand
            self.__analyse()

    def __reset(self):

        self.__lstAnalysis = None
        self.__errorFound = False
        self.__strCommand = None

    def __bool__(self):
        return not self.__errorFound

    def __analyse(self):

        strCommand = _convertToBashStyle(
            self.__strCommand
        )  # Comvert line to bash style

        self.__bashCommand = strCommand

        self.__lstAnalysis = []

        rg = r"^(.*?)\s\-\-.*?\-\-output.(.*?)\s\-\-.*?\s'\('\s(.*?)\s'\)'.*?\-\-track-order\s(.*)"

        regCommandEx = re.compile(rg)
        matchCommand = regCommandEx.match(strCommand)

        reExecutableEx = re.compile(r"^(.*?)\s\-\-")
        matchExecutable = reExecutableEx.match(strCommand)

        reOutputFileEx = re.compile(r".*?\-\-output\s(.*?)\s\-\-")
        matchOutputFile = reOutputFileEx.match(strCommand)

        reSourcesEx = re.compile(r"'\('\s(.*?)\s'\)'")
        matchSources = reSourcesEx.finditer(strCommand)

        reChaptersFileEx = re.compile(r".*?\-\-chapters\s(.*?)\s\-\-")
        matchChaptersFile = reChaptersFileEx.match(strCommand)

        reAttachmentsEx = re.compile(r"\-\-attach-file.(.*?)\s\-\-")
        matchAttachments = reAttachmentsEx.finditer(strCommand)

        bOk = True
        trackOrder = None

        # To look Ok must match the 5 group in the command line that
        # are expected
        # 1: mkvmerge name with fullpath
        # 2: output file
        # 3: at list one source
        # 4: track order
        if matchCommand and (len(matchCommand.groups()) == 4):
            self.__lstAnalysis.append("chk: Command seems ok.")
            trackOrder = matchCommand.group(4)
        else:
            self.__lstAnalysis.append("err: Command bad format.")
            bOk = False

        if trackOrder is not None:
            try:
                d = ast.literal_eval("{" + trackOrder + "}")
                trackTotal = _numberOfTracksInCommand(strCommand)

                s = trackOrder.split(",")
                if trackTotal == len(s):
                    for e in s:
                        if not e.find(":") > 0:
                            bOk = False
                else:
                    bOk = False

                if not bOk:
                    self.__lstAnalysis.append(
                        "err: Number of tracks {} and track order of {} don't match.".format(
                            trackTotal, len(d)
                        )
                    )
            except SyntaxError:
                self.__lstAnalysis.append("err: Command track order bad format.")
                bOk = False

        if matchExecutable:
            f = stripEncaseQuotes(matchExecutable.group(1))
            p = Path(f)
            try:
                test = p.is_file()
            except OSError:
                self.__lstAnalysis.append(
                    "err: mkvmerge incorrect syntax: - {}.".format(str(p))
                )
                bOk = False
            else:
                if test:
                    self.__lstAnalysis.append("chk: mkvmerge ok - {}".format(str(p)))
                else:
                    self.__lstAnalysis.append(
                        "err: mkvmerge not found - {}.".format(str(p))
                    )
                    bOk = False
        else:
            self.__lstAnalysis.append("err: mkvmerge not found.")
            bOk = False

        if matchOutputFile:
            f = stripEncaseQuotes(matchOutputFile.group(1))
            f = f.replace(r"'\''", "'")
            p = Path(f)
            self.__outputFile = None

            try:
                test = Path(p.parent).is_dir()
            except OSError:
                self.__lstAnalysis.append(
                    "err: Destination directory incorrect syntax - {}.".format(str(p.parent))
                )
                bOk = False
            else:
                if test:
                    self.__lstAnalysis.append(
                        "chk: Destination directory ok = {}".format(str(p.parent))
                    )
                    self.__outputFile = p
                else:
                    self.__lstAnalysis.append(
                        "err: Destination directory not found - {}.".format(str(p.parent))
                    )
                    bOk = False
        else:
            self.__lstAnalysis.append("err: Destination directory not found.")
            bOk = False

        if matchSources:
            n = 1

            for match in matchSources:
                f = _unQuote(match.group(1))
                p = Path(f)

                try:
                    test = Path(p.parent).is_dir()
                except OSError:
                    self.__lstAnalysis.append(
                        "err: Source directory {} bad syntax {}".format(n, str(p.parent))
                    )
                    bOk = False
                else:
                    if not test:
                        self.__lstAnalysis.append(
                            "err: Source directory {} not found {}".format(n, str(p.parent))
                        )
                        bOk = False
                    else:
                        self.__lstAnalysis.append(
                            "chk: Source directory {} ok = {}".format(n, str(p.parent))
                        )
                try:
                    test = Path(p).is_file()
                except OSError:
                    self.__lstAnalysis.append(
                        "err: Source file {} bad syntax {}".format(n, str(p))
                    )
                    bOk = False
                else:
                    if not test:
                        self.__lstAnalysis.append(
                            "err: Source file {} not found {}".format(n, str(p))
                        )
                        bOk = False
                    else:
                        self.__lstAnalysis.append(
                            "chk: Source file {} ok = {}".format(n, str(p))
                        )

                n += 1

            if n == 1:
                # if the command is so bad matchSources for loop won't run
                self.__lstAnalysis.append("err: Source directory not found.")
                bOk = False
        else:
            self.__lstAnalysis.append("err: Source directory not found.")
            bOk = False

        # Check for optional chapters file
        if matchChaptersFile:
            f = _unQuote(matchChaptersFile.group(1))
            p = Path(f)
            self.__chaptersFile = None

            try:
                test = p.is_file()
            except OSError:
                self.__lstAnalysis.append(
                    "err: Chapters file incorrect syntax - {}.".format(str(p))
                )
                bOk = False
            else:
                if not test:
                    self.__lstAnalysis.append(
                        "err: Chapters file not found - {}.".format(str(p))
                    )
                    bOk = False
                else:
                    self.__lstAnalysis.append("chk: Chapters file ok - {}".format(str(p)))
                    self.__chaptersFile = p

        # This check if for optional attachments files
        n = 1
        for match in matchAttachments:
            f = _unQuote(match.group(1))
            p = Path(p)
            try:
                test = p.is_file()
            except OSError:
                self.__lstAnalysis.append(
                    "err: Attachment file {} incorrect syntax - {}".format(n, str(p))
                )
                bOk = False
            else:
                if not test:
                    self.__lstAnalysis.append(
                        "err: Attachment {} not found - {}".format(n, str(p))
                    )
                    bOk = False
                else:
                    self.__lstAnalysis.append(
                        "chk: Attachment {} ok = {}".format(n, str(p))
                    )
            n += 1

        self.__errorFound = not bOk

        if self.log:

            for line in self.__lstAnalysis:
                if line.find("chk:") >= 0:
                    MODULELOG.debug("VMC0001: %s", line)
                elif line.find("err:") >= 0:
                    MODULELOG.error("VMC0001: %s", line)

    @property
    def log(self):
        """
        class property can be used to override the class global
        logging setting if set to None class log will be followed

        Returns:
            bool:

            True if logging is enable False otherwise
        """
        if self.__log is not None:
            return self.__log

        return VerifyMKVCommand.classLog()

    @log.setter
    def log(self, value):
        """set instance log variable"""
        if isinstance(value, bool) or value is None:
            self.__log = value

    @property
    def analysis(self):
        """
        results of analysis of the command

        Returns:
            list:

            list with comments of anything found
        """
        return self.__lstAnalysis

    @property
    def bashCommand(self):
        """
        windows command line converted to bash style

        Returns:
            str:

            command line for bash shell
        """
        return self.__bashCommand

    @property
    def command(self):
        """
        command line as generated by mkvtoolnix-gui

        Returns:
            str:

            current command line set
        """
        return self.__strCommand

    @command.setter
    def command(self, value):
        if isinstance(value, str):
            self.__reset()
            self.__strCommand = value
            self.__analyse()

    @property
    def outputFile(self):
        """
        Returns:
            pathlib.Path:

            cli output file pathlib.Path
        """
        return self.__outputFile

    @property
    def chaptersFile(self):
        """
        Returns:
            pathlib.Path:

            cli chapters file pathlib.Path
        """
        return self.__chaptersFile


def _convertToBashStyle(strCommand):
    """
    Strip escape windows chars for the command line
    in the end they won't be used in a shell
    the resulting command is bash/zh like

    Args:
        strCommand (str): command generated by mkvtoolnix-gui

    Returns:
        str:

        cli command converted to bash style
    """

    strTmp = strCommand

    if strTmp.find(r'^"^(^"') >= 0:
        # This is for cmd in Windows
        strTmp = (
            strTmp.replace("'", r"'\''")
            .replace("^", "")
            .replace("/", "\\")
            .replace('"', "'")
        )

    return strTmp


def _numberOfTracksInCommand(strCmd):
    """
    Every track have a --language option count
    them to know the number of tracks

    Args:
        strCmd(str): command line

    Returns:
        int:

        total number of tracks
    """

    reLanguageEx = re.compile(r"\-\-language (.*?)\s")
    matchLanguage = reLanguageEx.findall(strCmd)

    return len(matchLanguage)


def _unQuote(fileName):
    """
    Remove start end quotes and escape ones
    Args:
        fileName (str): file name

    Returns:
        str:

        file name without quotes if found
    """

    f = stripEncaseQuotes(fileName)
    f = f.replace(r"'\''", "'")

    return f



"""
process namespace
"""

from .classes import (
    GenericThreadWorker,
    ProcessWorker,
    QueueProcessWorker,
    QueueThreadWorker,
    RunCommand,
    ThreadWorker
)

from .utils import isThreadRunning



"""
Multithreading Class base on QThread
"""


import logging
import traceback
import sys


from PySide2.QtCore import QThread, QRunnable, Signal, Slot


MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class QthThread(QThread):
    """
    QThread generic class send function and arguments to start in thread

    :param function: The function callback to run on this worker thread. Supplied args and
                     kwargs will be passed through to the runner.
    :type function: function
    :param args: Arguments to pass to the callback function
    :param kwargs: Keywords to pass to the callback function
    """

    log = False

    def __init__(self, function, *args, **kwargs):
        super().__init__(self)

        self.function = function
        self.args = args
        self.kwargs = kwargs

    def __del__(self):
        self.wait()

    def run(self):
        """Override run and start function from argument"""

        self.function(*self.args, **self.kwargs)

        return


class QthThreadWorker(QRunnable):
    """
    Worker thread

    Inherits from threading.Tread to handle worker thread setup, signals and wrap-up.

    Args:
        function (function): Function to submit to Thread.
        funcFinished (function): Call back function when thread finishes.
        funcError (function): Call back function when an error occurs.
        funcResult (function): Call back function with the result of the execution.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    log = False

    def __init__(
        self,
        function,
        *args,
        funcFinished=None,
        funcError=None,
        funcResult=None,
        **kwargs
    ):
        super().__init__()

        # Store constructor arguments (re-used for processing)
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.finished = funcFinished
        self.error = funcError
        self.result = funcResult

    def run(self):
        """
        Override run initialise and starts the worker function
        with passed args, kwargs.
        """

        # Retrieve args/kwargs here; and fire processing using them
        # pylint: disable-msg=W0702
        # have to capture all exceptions using sys.exc_info()
        # to sort out what is happening
        try:
            result = self.function(*self.args, **self.kwargs)
        except:
            traceback.print_exc()
            excepttype, value = sys.exc_info()[:2]
            if callable(self.error):
                self.error((excepttype, value, traceback.format_exc()))
        else:
            if callable(self.result):
                self.result(result)  # Return the result of the processing
        finally:
            if callable(self.finished):
                self.finished()  # Done



"""
 SystemTrayIcon utility class
"""
import platform

from PySide2.QtCore import Slot
from PySide2.QtWidgets import QAction, QApplication, QMenu, QSystemTrayIcon
from PySide2 import QtWidgets

class QSystemTrayIconWidget(QSystemTrayIcon):
    """
    QSystemTrayIconWidget System tray icon utility class

    Args:
        icon (QIcon): icon to use
    """

    def __init__(self, parent, icon, seconds=5):
        super().__init__(parent)

        self.__isTrayIcon = False
        if QSystemTrayIcon.isSystemTrayAvailable():
            self.__isTrayIcon = True

            self.parent = parent
            self.icon = icon
            self.__seconds = seconds

            self.minimizeAction = QAction()
            self.maximizeAction = QAction()
            self.restoreAction = QAction()
            self.quitAction = QAction()

            self._trayIconActionsCreate()
            self._trayIconCreate(icon)

            self.activated.connect(self.iconActivated)

    def _trayIconActionsCreate(self):
        self.minimizeAction = QAction("Minimize", self.parent)
        self.minimizeAction.triggered.connect(self.parent.hide)

        self.maximizeAction = QAction("Maximize", self)
        self.maximizeAction.triggered.connect(self.parent.showMaximized)

        self.restoreAction = QAction("Restore", self.parent)
        self.restoreAction.triggered.connect(self.parent.showNormal)

        self.quitAction = QAction("Quit", self.parent)
        self.quitAction.triggered.connect(QApplication.quit)

    def _trayIconCreate(self, icon):

        if QSystemTrayIcon.isSystemTrayAvailable():
            self.__isTrayIcon = True

            trayIconMenu = QMenu(self.parent)
            trayIconMenu.addAction(self.minimizeAction)
            trayIconMenu.addAction(self.maximizeAction)
            trayIconMenu.addAction(self.restoreAction)
            trayIconMenu.addSeparator()
            trayIconMenu.addAction(self.quitAction)

            if platform.system() == "Windows":
                style = QtWidgets.QStyleFactory.create("windowsvista")
                trayIconMenu.setStyle(style)

            self.setContextMenu(trayIconMenu)
            self.setIcon(icon)

    def __bool__(self):
        return self.__isTrayIcon

    @property
    def isTrayIcon(self):
        return self.__isTrayIcon

    @property
    def seconds(self):
        return self.__seconds

    @seconds.setter
    def seconds(self, value):
        if isinstance(value, int):
            self.__seconds = value

    @Slot(str)
    def iconActivated(self, reason):
        print("iconActivated")
        if reason == QSystemTrayIcon.Trigger:
            print("Trigger")
        if reason == QSystemTrayIcon.DoubleClick:
            print("DoubleClick")
        if reason == QSystemTrayIcon.MiddleClick:
            print("MiddleClick")
        if reason == QSystemTrayIcon.Context:
            print("Context Request")

    @Slot(int)
    def setVisible(self, visible):
        print("Visible")
        super().visible(visible)

    @Slot(int)
    def setMenuEnabled(self, visible):
        """ Override setVisible """
        if self.__isTrayIcon:
            self.minimizeAction.setEnabled(visible)
            self.maximizeAction.setEnabled(not self.parent.isMaximized())
            self.restoreAction.setEnabled(self.parent.isMaximized() or not visible)

    @Slot(str, str)
    @Slot(str, str, object)
    @Slot(str, str, object, int)
    def showMessage(self, title, msg, icon=QSystemTrayIcon.Information, mSecs=1000):

        super().showMessage(title, msg, icon, self.__seconds + 1000)



"""
Import names for process
"""

# THR0001

from .multithreading import GenericThreadWorker, QueueThreadWorker, ThreadWorker
from .multiprocessing import ProcessWorker, QueueProcessWorker
from .runcommand import RunCommand



"""Utility functions"""


import ctypes
import platform


def isMediaInfoLib():
    """
    Check if MediaInfo library is on the system

    Returns:
        bool: True if found False otherwise
    """

    currentOS = platform.system()
    libNames = ()

    if currentOS == "Windows":
        libraryType = ctypes.WinDLL
        libNames = ["MediaInfo.dll"]
    else:
        libraryType = ctypes.CDLL

    if currentOS == "Darwin":
        libNames = ["libmediainfo.0.dylib", "libmediainfo.dylib"]
    elif currentOS == "Linux":
        libNames = ["libmediainfo.so.0", "libzen.so.0"]

    for library in libNames:
        try:
            libFile = libraryType(library) # pylint: disable=W0612
        except OSError:
            return False

    return True



"""
format timedelta object
"""

from string import Formatter


def strFormatTimeDelta(tDelta, fmt=None):
    """
    strFormatTimeDelta format time delta string. If fmt is None a format of
    'N days N hours N minutes N seconds' will be used starting with the first
    non zero N found in the calculations.


    Args:
        tDelta (timedelta): timedelta object to format
        fmt (str, optional): string formatter. Defaults to None.

    Returns:
        str: formatted timedelta string
    """

    strFormatter = Formatter()
    args = {}
    divisor = {'W': 604800, 'D': 86400, 'H': 3600, 'M': 60, 'S': 1}
    labels = {'W': "weeks", 'D': "days", 'H': "hours", 'M': "minutes", 'S': "seconds"}

    remainder = int(tDelta.total_seconds())

    if fmt is not None:
        keywords = list(map(lambda x: x[1], list(strFormatter.parse(fmt))))
        for key in divisor:
            if key in keywords and key in divisor.keys():
                args[key], remainder = divmod(remainder, divisor[key])
    else:
        firstFound = False
        strTmp = ""
        for key in divisor:
            quotient, remainder = divmod(remainder, divisor[key])
            if firstFound:
                strTmp += " " + str(quotient) + " " + labels[key]
            elif quotient > 0:
                strTmp += str(quotient) + " " + labels[key]
                firstFound = True

    if fmt is None:
        return strTmp

    return strFormatter.format(fmt, **args)



"""
macOS convenience functions
"""

import platform
import sys

from vsutillib.process import RunCommand


def isMacDarkMode():
    """
    Test for macOS Mojave Dark Mode

    Returns:
        bool:

        True if the macOS is using Dark Mode

        False if not or if called on other operating systems
    """

    if platform.system() == "Darwin":
        cmd = RunCommand("defaults read -g AppleInterfaceStyle")

        if getattr(sys, 'frozen', False):
            # running in pyinstaller bundle dark mode does not apply
            return False

        if cmd.run():
            for e in cmd.output:
                if e.find("Dark") >= 0:
                    return True

    return False



"""
 TabWidgetExtension
"""

class TabWidgetExtension:

    def __init__(self, tabWidgetChild=None, tabWidget=None, **kwargs):
        super().__init__(**kwargs)

        self.__tabWidgetChild = tabWidgetChild
        self.__tabWidget = tabWidget
        self.__tab = None
        self.__oldTab = (-1)
        self.__title = None
        self.__toolTip = None

    @property
    def tab(self):
        return self.__tab

    @tab.setter
    def tab(self, value):
        self.__tab = value

    @property
    def tabWidget(self):
        return self.__tabWidget

    @tabWidget.setter
    def tabWidget(self, value):
        self.__tabWidget = value

    @property
    def title(self):
        return self.__title

    @title.setter
    def title(self, value):
        self.__title = value
        if self.tab >= 0:
            self.tabWidget.setTabText(self.tab, self.title)

    @property
    def toolTip(self):
        return self.__toolTip

    @toolTip.setter
    def toolTip(self, value):
        self.__toolTip = value
        if self.tab >= 0:
            self.tabWidget.setTabToolTip(self.tab, self.toolTip)

    def setAsCurrentTab(self):
        if self.tab >= 0:
            self.tabWidget.setCurrentIndex(self.tab)

    def hideTab(self):

        tabIndex = self.tabWidget.indexOf(self.__tabWidgetChild)
        self.__oldTab = tabIndex
        self.tabWidget.removeTab(tabIndex)
        self.tab = (-1)

    def unHideTab(self):

        if self.tabWidget is not None:
            if self.__oldTab >= 0:
                tabIndex = self.tabWidget.insertTab(self.__oldTab, self.__tabWidgetChild, self.title)
            else:
                tabIndex = self.tabWidget.addTab(self.__tabWidgetChild, self.title)
            self.tabWidget.setTabToolTip(tabIndex, self.toolTip)
            self.tab = tabIndex



"""
PySide2 related classes
"""

from .insertTextHelpers import LineOutput, checkColor
from .QActionWidget import QActionWidget
from .QComboLineEdit import QComboLineEdit
from .QFormatLabel import QFormatLabel
from .QFileListWidget import QFileListWidget
from .QGenericThreads import QthThread, QthThreadWorker
from .QGroupSignal import QGroupSignal
from .QLineFrame import HorizontalLine, VerticalLine
from .QMenuWidget import QMenuWidget
from .QOutputTextWidget import QOutputTextWidget
from .QProgressBar import DualProgressBar, TaskbarButtonProgress
from .QProgressIndicator import QProgressIndicator
from .QPushButtonWidget import QPushButtonWidget
from .QRunInThread import QRunInThread
from .QSignalLogHandler import QSignalLogHandler
from .QSystemTrayIconWidget import QSystemTrayIconWidget
from .QWorker import Worker, WorkerSignals
from .SvgColor import SvgColor
from .TabWidget import TabWidget
from .TabWidgetExtension import TabWidgetExtension



"""
RunCommand

Run a command in a subprocess and capture any output

if processLine function is provided it will be called
with every line read

if regexsearch regular expresion is provided the first
match will be set on regexmatch property
"""
# RNC0001
# pylint: disable=bad-continuation

import logging
import platform
import re
import shlex
import subprocess
import traceback

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class RunCommand:
    """
    Run a command in a subprocess and capture any output

    processLine function if provided it will be called
    with every line read.

    regexsearch regular expresion if provided the first
    match will be set on regexmatch property

    Args:
        command (str): command to execute
        commandShlex (:obj:`bool`): True if command is shlex.split
            False otherwise. Defaults to False.
        processLine (:obj:`function`, optional): Function called with
            every line read. Defaults to None.
        processArgs (:obj:`list`, optional): Variable length list to
            pass to processLine. Defaults to None.
        processKWArgs (:obj:`list`, optional): Arbitrary keyword
            arguments to pass to processLine. Defaults to None.
        regexsearch (:obj:`str`, optional): Regex applied to every
            line read. Defaults to None
        universalNewLine (:obj:`bool`): True to read in text mode
            False to read binary mode. Defaults to False.

    Raises:

        ValueError: If processArgs is not a list or if processKWArgs
            is not a dictionary.
    """

    __log = False

    @classmethod
    def classLog(cls, setLogging=None):
        """
        get/set logging at class level
        every class instance will log
        unless overwritten

        Args:
            setLogging (`bool`):

                - True class will log
                - False turn off logging
                - None returns current Value

        Returns:
            bool:

            returns the current value set
        """

        if setLogging is not None:
            if isinstance(setLogging, bool):
                cls.__log = setLogging

    def __init__(
        self,
        command=None,
        processLine=None,
        processArgs=None,
        processKWArgs=None,
        regexsearch=None,
        commandShlex=False,
        universalNewLines=False,
        controlQueue=None,
        log=None,
    ):  # pylint: disable=too-many-arguments

        self.__command = None
        self.command = command  # Call class setter property

        self.__commandShlex = commandShlex
        self.__processLine = processLine
        self.__universalNewLines = universalNewLines
        self.__processArgs = []

        if processArgs is not None:
            if isinstance(processArgs, list):
                self.__processArgs = processArgs
            else:
                raise ValueError("processLineParam has to be a list")

        self.__processKWArgs = {}

        if processKWArgs is not None:
            if isinstance(processKWArgs, dict):
                self.__processKWArgs = processKWArgs
            else:
                raise ValueError("processLineParam has to be a dictionary")

        self.__regEx = None
        if regexsearch is not None:
            if isinstance(regexsearch, list):
                self.__regEx = []
                for regex in regexsearch:
                    self.__regEx.append(re.compile(regex))
            else:
                self.__regEx = re.compile(regexsearch)

        self.__error = ""
        self.__output = []
        self.__controlQueue = controlQueue
        self.__returnCode = None
        self.__regexmatch = None
        self.__log = None

    def __bool__(self):
        if self.__command:
            return True
        return False

    @property
    def log(self):
        """
        class property can be used to override the class global
        logging setting if set to None class log will be followed

        Returns:
            bool:

            True if logging is enable False otherwise
        """
        if self.__log is not None:
            return self.__log

        return RunCommand.classLog()

    @log.setter
    def log(self, value):
        """set instance log variable"""
        if isinstance(value, bool) or value is None:
            self.__log = value

    @property
    def command(self):
        """return current command set in class"""

        return self.__command

    @command.setter
    def command(self, value):
        """
        command to execute

        Args:
            command (str): command to execute

        Returns:
            str:

            current command set
        """
        self.__command = None
        self._reset(value)

    @property
    def shlexCommand(self):
        """
        command to submit to subproccess PIPE

        Returns:
            list:

            command split by shlex.split
        """
        return shlex.split(self.__command)

    @property
    def error(self):
        """
        error if command can not be executed

        Returns:
            str:

            message if command fails to execute
        """
        return self.__error

    @property
    def output(self):
        """
        captured output

        Returns:
            list:

            output of executed command
        """
        return self.__output

    @property
    def parsedcommand(self):
        """
        command parsed by shlex
        can be used for debugging

        Returns:
            list|dict:

            depending of the regex returns a list or
            re.Match object
        """
        return shlex.split(self.__command)

    @property
    def rc(self):
        """
        Return code. On Windows is not reliable information.

        Returns:
            int:

            return code of executed command
        """
        return self.__returnCode

    @property
    def regexmatch(self):
        """
        results of regular expresion search

        Returns:
            list|dict:

            list if matches if single regex passed.  dict of list
            with the regex as key if a list of regex is passed.
        """
        return self.__regexmatch

    def run(self):
        """
        method to submit command to subprocess PIPE
        """

        self._reset()
        self._getCommandOutput()
        if self.__output:
            return True
        return False

    def _reset(self, command=None):
        """reset internal variables"""

        self.__output = []
        self.__error = ""
        self.__regexmatch = None
        if command is not None:
            self.__command = command

    def _regexMatch(self, line):
        """Have to set the size of in case of list"""

        if isinstance(self.__regEx, list):
            for index, regex in enumerate(self.__regEx):
                if m := regex.search(line):
                    if self.__regexmatch is None:
                        self.__regexmatch = [None] * len(self.__regEx)
                    tmpList = []
                    for i in m.groups():
                        tmpList.append(i)
                    self.__regexmatch[index] = tmpList
        else:
            if self.__regEx:
                if m := self.__regEx.search(line):
                    if self.__regexmatch is None:
                        self.__regexmatch = []
                    for i in m.groups():
                        self.__regexmatch.append(i)

    def _getCommandOutput(self):
        """Execute command in a subprocess"""

        self.__returnCode = 10000
        rc = 1000
        if self.__commandShlex:
            cmd = self.__command
        else:
            cmd = shlex.split(self.__command)
        try:
            if platform.system() == "Windows":
                p = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    universal_newlines=self.__universalNewLines,
                    stderr=subprocess.STDOUT,
                    creationflags=subprocess.CREATE_NO_WINDOW,
                )
            else:
                p = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    universal_newlines=self.__universalNewLines,
                    stderr=subprocess.STDOUT,
                )
            try:
                for l in p.stdout:
                    if self.__universalNewLines:
                        line = l
                    else:
                        line = l.decode("utf-8")
                    self.__output.append(line)
                    self._regexMatch(line)
                    if self.__processLine is not None:
                        self.__processLine(
                            line, *self.__processArgs, **self.__processKWArgs
                        )
                    if self.__controlQueue:
                        queueStatus = self.__controlQueue.popleft()
                        self.__controlQueue.appendleft(queueStatus)
                        if queueStatus in [
                            RunStatus.Abort,
                            RunStatus.AbortJob,
                            RunStatus.AbortForced,
                        ]:
                            # print("Aborting = {}".format(queueStatus))
                            p.kill()
                            outs, errs = p.communicate()
                            # if outs:
                            #    print(outs)
                            # if errs:
                            #    print(errs)
                            # print("rc = ", p.returncode)
                            rc = p.returncode
                            self.__returnCode = p.returncode
                            break
            except UnicodeDecodeError as error:
                trb = traceback.format_exc()
                msg = "Error: {}".format(error.reason)
                self.__output.append(str(cmd) + "\n")
                self.__output.append(msg)
                self.__output.append(trb)
                if self.__processLine is not None:
                    self.__processLine(
                        line, *self.__processArgs, **self.__processKWArgs
                    )
                if self.log:
                    MODULELOG.debug("RNC0001: Unicode decode error %s", msg)
            except KeyboardInterrupt as error:
                trb = traceback.format_exc()
                msg = "Error: {}".format(error.args)
                self.__output.append(str(cmd) + "\n")
                self.__output.append(msg)
                self.__output.append(trb)
                if self.__processLine is not None:
                    self.__processLine(
                        line, *self.__processArgs, **self.__processKWArgs
                    )
                if self.log:
                    MODULELOG.debug("RNC0002: Keyboard interrupt %s", msg)
                raise SystemExit(0)
            if rcResult := p.poll():
                self.__returnCode = rcResult
                rc = rcResult
        except FileNotFoundError as e:
            self.__error = e

        return rc

    def _getCommandOutputBackupTwo(self):
        """Execute command in a subprocess"""

        self.__returnCode = 10000
        rc = 1000
        if self.__commandShlex:
            cmd = self.__command
        else:
            cmd = shlex.split(self.__command)
        try:
            with subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                universal_newlines=self.__universalNewLines,
                stderr=subprocess.STDOUT,
                creationflags=0x08000000,
            ) as p:
                try:
                    for l in p.stdout:
                        if self.__universalNewLines:
                            line = l
                        else:
                            line = l.decode("utf-8")
                        self.__output.append(line)
                        self._regexMatch(line)
                        if self.__processLine is not None:
                            self.__processLine(
                                line, *self.__processArgs, **self.__processKWArgs
                            )
                        if self.__controlQueue:
                            queueStatus = self.__controlQueue.popleft()
                            self.__controlQueue.appendleft(queueStatus)
                            if queueStatus in [
                                RunStatus.Abort,
                                RunStatus.AbortJob,
                                RunStatus.AbortForced,
                            ]:
                                # print("Aborting = {}".format(queueStatus))
                                p.kill()
                                outs, errs = p.communicate()
                                # if outs:
                                #    print(outs)
                                # if errs:
                                #    print(errs)
                                # print("rc = ", p.returncode)
                                rc = p.returncode
                                self.__returnCode = p.returncode
                                break
                except UnicodeDecodeError as error:
                    trb = traceback.format_exc()
                    msg = "Error: {}".format(error.reason)
                    self.__output.append(str(cmd) + "\n")
                    self.__output.append(msg)
                    self.__output.append(trb)
                    if self.__processLine is not None:
                        self.__processLine(
                            line, *self.__processArgs, **self.__processKWArgs
                        )
                    if self.log:
                        MODULELOG.debug("RNC0001: Unicode decode error %s", msg)
                except KeyboardInterrupt as error:
                    trb = traceback.format_exc()
                    msg = "Error: {}".format(error.args)
                    self.__output.append(str(cmd) + "\n")
                    self.__output.append(msg)
                    self.__output.append(trb)
                    if self.__processLine is not None:
                        self.__processLine(
                            line, *self.__processArgs, **self.__processKWArgs
                        )
                    if self.log:
                        MODULELOG.debug("RNC0002: Keyboard interrupt %s", msg)
                    raise SystemExit(0)
                if rcResult := p.poll():
                    self.__returnCode = rcResult
                    rc = rcResult
                # print("Return Code = {}".format(rc))
        except FileNotFoundError as e:
            self.__error = e

        return rc

    def _getCommandOutputBackup(self):
        """Execute command in a subprocess"""

        self.__returnCode = 10000
        rc = 1000
        if self.__commandShlex:
            cmd = self.__command
        else:
            cmd = shlex.split(self.__command)
        try:
            with subprocess.Popen(
                cmd, stdout=subprocess.PIPE, bufsize=1, stderr=subprocess.STDOUT
            ) as p:
                try:
                    for l in iter(p.stdout):
                        line = l.decode("utf-8")
                        self.__output.append(line)
                        self._regexMatch(line)
                        if self.__processLine is not None:
                            self.__processLine(
                                line, *self.__processArgs, **self.__processKWArgs
                            )
                except UnicodeDecodeError as error:
                    trb = traceback.format_exc()
                    msg = "Error: {}".format(error.reason)
                    self.__output.append(str(cmd) + "\n")
                    self.__output.append(msg)
                    self.__output.append(trb)
                    if self.__processLine is not None:
                        self.__processLine(line)
                except KeyboardInterrupt as error:
                    trb = traceback.format_exc()
                    msg = "Error: {}".format(error.args)
                    self.__output.append(str(cmd) + "\n")
                    self.__output.append(msg)
                    self.__output.append(trb)
                    if self.__processLine is not None:
                        self.__processLine(line)
                    raise SystemExit(0)
                rcResult = p.poll()
                if rcResult is not None:
                    self.__returnCode = rcResult
                    rc = rcResult
        except FileNotFoundError as e:
            self.__error = e

        return rc


class RunStatus:
    """Key values for job related work"""

    Abort = "Abort"
    Aborted = "Aborted"
    AbortForced = "AbortForced"
    AbortJob = "AbortJob"
    AbortJobError = "AbortJobError"
    AddToQueue = "AddToQueue"
    Blocked = "Blocked"
    Done = "Done"
    DoneWithError = "DoneWithError"
    Error = "Error"
    Queue = "Queue"
    Running = "Running"
    Skip = "Skip"
    Skipped = "Skipped"
    Stop = "Stop"
    Stopped = "Stopped"
    Waiting = "Waiting"



"""
qtUtils:

utility functions that use PySide2
"""


import logging
import platform

from PySide2.QtCore import Qt
from PySide2.QtGui import QPalette, QColor
from PySide2.QtWidgets import QCheckBox, QDesktopWidget, QPushButton, QToolTip

from vsutillib.macos import isMacDarkMode

from .classes import QRunInThread, SvgColor

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


def centerWidget(widget, parent=None):
    """center widget based on parent or screen geometry"""

    if parent is None:
        parent = widget.parentWidget()

    if parent:
        widget.move(parent.frameGeometry().center() - widget.frameGeometry().center())

    else:
        widget.move(
            QDesktopWidget().availableGeometry().center()
            - widget.frameGeometry().center()
        )


def pushButton(label, function, tooltip):
    """
    pushButton convenience function for QPushButton definitions

    Args:
        label (str): label for push button
        function (func): function to execute when clicked
        tooltip (str): tool tip for push button

    Returns:
        QPushButton: PySide2 QPushButton
    """

    button = QPushButton(label)
    button.resize(button.sizeHint())
    button.clicked.connect(function)
    button.setToolTip(tooltip)

    return button


def darkPalette(app=None):

    palette = QPalette()

    darkColor = QColor(45, 45, 45)
    disabledColor = QColor(127, 127, 127)

    palette.setColor(QPalette.AlternateBase, darkColor)
    palette.setColor(QPalette.Base, QColor(18, 18, 18))
    palette.setColor(QPalette.BrightText, SvgColor.red)
    palette.setColor(QPalette.Button, darkColor)
    palette.setColor(QPalette.ButtonText, SvgColor.white)
    palette.setColor(QPalette.Disabled, QPalette.ButtonText, disabledColor)
    palette.setColor(QPalette.Disabled, QPalette.HighlightedText, disabledColor)
    palette.setColor(QPalette.Disabled, QPalette.Text, disabledColor)
    palette.setColor(QPalette.Disabled, QPalette.WindowText, disabledColor)
    palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
    palette.setColor(QPalette.HighlightedText, SvgColor.black)
    palette.setColor(QPalette.Link, QColor(42, 130, 218))
    palette.setColor(QPalette.Text, SvgColor.white)
    palette.setColor(QPalette.ToolTipBase, SvgColor.white)
    palette.setColor(QPalette.ToolTipText, SvgColor.white)
    palette.setColor(QPalette.Window, darkColor)
    palette.setColor(QPalette.WindowText, SvgColor.white)

    if app is not None:
        app.setStyle("Fusion")
        #app.setStyle("windowsvista")
        app.setPalette(palette)

        toolTipPalette = QPalette()
        toolTipPalette.setColor(QPalette.Inactive, QPalette.ToolTipText, SvgColor.white)
        toolTipPalette.setColor(
            QPalette.Inactive, QPalette.ToolTipBase, QColor(42, 130, 218)
        )
        QToolTip.setPalette(toolTipPalette)

    return palette

def qtRunFunctionInThread(function, *args, **kwargs):
    """
    Pass the function to execute Other args,
    kwargs are passed to the run function
    """

    funcStart = kwargs.pop("funcStart", None)
    funcFinished = kwargs.pop("funcFinished", None)
    funcResult = kwargs.pop("funcResult", None)

    worker = QRunInThread(function, *args, **kwargs)

    if funcStart is not None:
        worker.startSignal.connect(funcStart)
    if funcFinished is not None:
        worker.finishedSignal.connect(funcFinished)
    if funcResult is not None:
        worker.resultSignal.connect(funcResult)

    # Execute
    worker.run()



"""
QDarkPalette class
"""

from PySide2.QtGui import QPalette, QColor


WHITE = QColor(255, 255, 255)
BLACK = QColor(0, 0, 0)
RED = QColor(255, 0, 0)
PRIMARY = QColor(53, 53, 53)
SECONDARY = QColor(35, 35, 35)
TERTIARY = QColor(42, 130, 218)


def css_rgb(color, a=False):
    """Get a CSS `rgb` or `rgba` string from a `QtGui.QColor`."""
    return ("rgba({}, {}, {}, {})" if a else "rgb({}, {}, {})").format(*color.getRgb())


class QDarkPalette(QPalette):
    """Dark palette for a Qt application meant to be used with the Fusion theme."""

    def __init__(self, *__args):
        super().__init__(*__args)

        # Set all the colors based on the constants in globals
        self.setColor(QPalette.Window, PRIMARY)
        self.setColor(QPalette.WindowText, WHITE)
        self.setColor(QPalette.Base, SECONDARY)
        self.setColor(QPalette.AlternateBase, PRIMARY)
        self.setColor(QPalette.ToolTipBase, WHITE)
        self.setColor(QPalette.ToolTipText, WHITE)
        self.setColor(QPalette.Text, WHITE)
        self.setColor(QPalette.Button, PRIMARY)
        self.setColor(QPalette.ButtonText, WHITE)
        self.setColor(QPalette.BrightText, RED)
        self.setColor(QPalette.Link, TERTIARY)
        self.setColor(QPalette.Highlight, TERTIARY)
        self.setColor(QPalette.HighlightedText, BLACK)

    @staticmethod
    def setStyleSheet(app):
        """Static method to set the tooltip stylesheet to a `QtWidgets.QApplication`."""
        app.setStyleSheet(
            "QToolTip {{"
            "color: {white};"
            "background-color: {tertiary};"
            "border: 1px solid {white};"
            "}}".format(white=css_rgb(WHITE), tertiary=css_rgb(TERTIARY))
        )

    def setA36pp(self, app):
        """Set the Fusion theme and this palette to a `QtWidgets.QApplication`."""
        app.setStyle("Fusion")
        app.setPalette(self)
        self.set_stylesheet(app)



"""
Job info widget

format is Jobs: {:3d} Current Job: {:3d} File: {:3d} of {:3d} Errors: {:3d}

Jobs: total jobs
jobID: current running job
file: file number been work on current job
totalFiles: total files on current job
errors: total errors on current job
align:
    Qt.Horizontal - Horizontal layout the default
    Qt.Vertical - Vertical layout
"""

import random

from PySide2.QtWidgets import QWidget, QLabel
from PySide2.QtCore import Slot, Signal


class QFormatLabel(QLabel):
    """
    Dual QProgressBar for unit and total progress

    param align - Set alignment Qt.Horizontal or Qt.Vertical
    type align - Qt.AlignmentFlags
    """

    setTemplateSignal = Signal(str)
    setValueSignal = Signal(int, object)
    setValuesSignal = Signal(list)

    def __init__(self, *args, **kwargs):

        template = kwargs.pop("template", None)
        initValues = kwargs.pop("init", None)

        if (args) or (template is not None):
            if initValues is None:
                raise KeyError("init= not specified")

            if args:
                if isinstance(args[0], str):
                    template = args[0]
                    args = args[1:]

        if initValues is None:
            initValues = []

        super().__init__(*args, **kwargs)

        self.setTemplateSignal.connect(self.setTemplate)

        if template is None:
            self._template = (
                "Job(s): {0:3d} Running: {1:3d} File: {2:3d} of {3:3d} Errors: {4:3d}"
            )
            initValues = [0, 0, 0, 0, 0]
        else:
            self._template = template

        self._values = initValues
        self._refresh()

    def __getitem__(self, index):
        return self._values[index]

    def __setitem__(self, index, value):
        self.setValueSignal.emit(index, value)
        #self._values[index] = value
        #self._refresh()

    def _refresh(self):

        strTmp = self._template
        strTmp = strTmp.format(*self._values)
        super().setText(strTmp)

    @property
    def template(self):
        return self._template

    @template.setter
    def template(self, value):

        if isinstance(value, str):
            self._template = value
            self._refresh()

    @Slot(str)
    def setTemplate(self, value):

        if isinstance(value, str):
            self._template = value
            self._refresh()

    @Slot(list)
    def setValues(self, args):
        """
        Set Values

        :param args: set values
        :type args: list
        """

        self._values = list(args)
        self._refresh()

    @Slot(int, object)
    def setValue(self, index, value):
        """
        Set value index based

        :param index: index position
        :type index: int
        :param value: value to set
        :type value: object
        """

        self._values[index] = value
        self._refresh()

    @property
    def values(self):
        """return current positional values"""

        return self._values

    def valuesConnect(self, signal):
        """make connection to setValues Slot"""

        signal.connect(self.setValues)

    def valueConnect(self, signal):
        """make connection to setValues Slot"""

        signal.connect(self.setValue)


if __name__ == "__main__":

    import sys
    from PySide2.QtWidgets import QApplication, QGridLayout, QMainWindow, QPushButton

    class MainWindow(QMainWindow):
        """Test the progress bars"""

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            l = QGridLayout()

            self.jobInfo = FormatLabel()
            self.formatLabel = FormatLabel(
                "Random 1 = {0:>3d} -- Random 2 = {0:3d}", init=[0, 0]
            )

            b = QPushButton("Test 1")
            b.pressed.connect(self.test)
            b1 = QPushButton("Test 2")
            b1.pressed.connect(self.test1)
            l.addWidget(self.jobInfo, 0, 0)
            l.addWidget(self.formatLabel, 1, 0)
            l.addWidget(b, 4, 0)
            l.addWidget(b1, 5, 0)

            w = QWidget()
            w.setLayout(l)

            self.setCentralWidget(w)

            self.show()

        def test(self):
            """Test"""

            r = random.randint

            self.jobInfo.setValues(
                (r(1, 1001), r(1, 1001), r(1, 1001), r(1, 1001), r(1, 1001))
            )

        def test1(self):
            """Test FormatLabel"""
            r = random.randint

            self.formatLabel[0] = r(1, 1001)
            self.formatLabel[1] = r(1, 1001)

    # pylint: disable=C0103
    # variables use to construct application not constants
    app = QApplication([])
    window = MainWindow()
    sys.exit(app.exec_())



"""
Convenience functions
"""

import threading


def isThreadRunning(threadName):
    """
    isThreadRunning check for named thread running

    Args:
        threadName (str): name of thread to look for

    Returns:
        bool: True if name found, False otherwise.
    """

    if threadName in [x.name for x in threading.enumerate()]:
        return True

    return False



"""
Sub-class of QTextEdit accepts drag and drop
"""
import logging

from pathlib import Path

from natsort import natsorted, ns

from PySide2.QtCore import Slot, Signal
from PySide2.QtWidgets import QMenu


from .QOutputTextWidget import QOutputTextWidget


MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class QFileListWidget(QOutputTextWidget):
    """
    QTextEdit subclass that accepts dropped files
    displays only the name of the files.
    The full path is save and use internally.
    """

    filesDroppedUpdateSignal = Signal(list)

    def __init__(self, parent):
        super().__init__(parent)

        # self.setDragEnabled(True)
        self.fileList = []
        self.bBlockDrops = False
        self.bFilesDropped = False

    def clear(self):
        self.fileList = []
        self.bBlockDrops = False
        self.bFilesDropped = False
        super().setAcceptDrops(True)
        super().clear()

    def dragEnterEvent(self, event):

        data = event.mimeData()
        urls = data.urls()
        if urls and urls[0].scheme() == "file":
            event.acceptProposedAction()

    def dragMoveEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if urls and urls[0].scheme() == "file":
            event.acceptProposedAction()

    def dropEvent(self, event):
        data = event.mimeData()
        urls = data.urls()

        bUpdate = False
        for f in urls:
            filePath = str(f.path())[1:]

            fPath = Path(filePath)

            if fPath.is_dir():
                files = [x for x in fPath.glob("*.*") if x.is_file()]
                files = natsorted(files, alg=ns.PATH)
                for x in files:
                    if x not in self.fileList:
                        self.fileList.append(x)
                        bUpdate = True
            elif fPath.is_file():
                if fPath not in self.fileList:
                    self.fileList.append(fPath)
                    bUpdate = True

        if bUpdate:
            if not self.bFilesDropped:
                self.bFilesDropped = True

            self._displayFiles()
            self.filesDroppedUpdateSignal.emit(self.fileList)

    def contextMenuEvent(self, event):

        if self.bFilesDropped and self.fileList:
            menu = QMenu(self)
            clearAction = menu.addAction(Actions.Clear)
            sortAction = menu.addAction(Actions.Sort)
            action = menu.exec_(self.mapToGlobal(event.pos()))
            if action == clearAction:
                self.clear()
                self.filesDroppedUpdateSignal.emit(self.fileList)
            elif action == sortAction:
                if self.fileList:
                    self.fileList.sort(key=str)
                    self._displayFiles()
                    self.filesDroppedUpdateSignal.emit(self.fileList)

    def setAcceptDrops(self, value):

        if not self.bBlockDrops:
            # don't check for type to raise error
            super().setAcceptDrops(value)

    @Slot(list)
    def setFileList(self, filesList=None):
        """
        Set the files manually

        Args:
            filesList (list, optional): file list to display. Defaults to None.
        """

        if filesList:
            self.bBlockDrops = True
            self.bFilesDropped = False
            super().setAcceptDrops(False)
            self.fileList = []

            for f in filesList:
                self.fileList.append(f)

            self._displayFiles()

    def _displayFiles(self):
        """display the files on QTextEdit box"""

        super().clear()

        for f in self.fileList:
            self.insertPlainText(f.name + "\n")


class Actions:
    """
    Actions labels for context menu
    """

    Clear = "Clear"
    Sort = "Sort"



"""
 Parse mkvmerge into a class

Track options assumed appear after --language

r"|--sync "
r"|--cues "
r"|--default-track "
r"|--forced-track "
r"|--blockadd "
r"|--track-name "
r"|--tags "
r"|--aac-is-sbr "
r"|--reduce-to-core "
r"|--remove-dialog-normalization-gain  "
r"|--timestamps "
r"|--default-duration "
r"|--fix-bitstream-timing-information "
r"|--nalu-size-length "
r"|--compression "


Properties update

<MKVPROPEDIT.EXE> <SOURCE> --edit info --set title=<TITLE>

"""
# MCP0003

import ast
import logging
import platform
import re
import shlex

from pathlib import Path

from vsutillib.media import MediaFileInfo
from vsutillib.misc import XLate

from ..mkvutils import (
    convertToBashStyle,
    numberOfTracksInCommand,
    resolveOverwrite,
    stripEncaseQuotes,
    unQuote,
)
from .mkvattachments import MKVAttachments
from .mkvclassutil import SourceFile, SourceFiles

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class MKVCommandParser:
    """
    Class for parsing CLI **mkvmerge** command part of MKVToolnix_

    Args:
        strCommand (`str`, optional): command line as generated by mkvtoolnix-gui
    """

    __log = False

    @classmethod
    def classLog(cls, setLogging=None):
        """
        get/set logging at class level
        every class instance will log
        unless overwritten

        Args:
            setLogging (bool):
                - True class will log
                - False turn off logging
                - None returns current Value

        Returns:
            bool:

            returns the current value set
        """

        if setLogging is not None:
            if isinstance(setLogging, bool):
                cls.__log = setLogging

        return cls.__log

    def __init__(self, strCommand=None, log=None):

        self.log = log
        self.command = strCommand

    def _initVars(self):

        self.__bashCommand = None
        self.__errorFound = False
        self.__log = None
        self.__lstAnalysis = None
        self.__totalSourceFiles = None
        self.__readFiles = False
        self.__strCommand = None
        self.__shellCommands = []
        self.__strCommands = []

        self.cliChaptersFile = None
        self.cliChaptersFileMatchString = None
        self.chaptersLanguage = None
        self.commandTemplate = None
        self.language = None
        self.mkvmerge = None
        self.mkvpropedit = None
        self.cliOutputFile = None
        self.cliOutputFileMatchString = None
        self.cliTitleMatchString = None
        self.cliTrackOrder = None

        self.chaptersFiles = []
        self.filesInDirByKey = {}
        self.dirsByKey = {}
        self.titles = []

        self.oAttachments = MKVAttachments()
        self.oSourceFiles = SourceFiles()

    def __bool__(self):
        return not self.__errorFound

    def __contains__(self, item):
        return item in self.__strCommands

    def __getitem__(self, index):
        return (
            self.__shellCommands[index],
            self.baseFiles,
            self.oSourceFiles[index],
            self.filesInDirByKey[MKVParseKey.outputFile][index],
            None if not self.oAttachments else self.oAttachments.attachmentsStr[index],
            None if not self.titles else self.titles[index],
            None if not self.cliChaptersFile else self.chaptersFiles[index],
        )

    def __len__(self):
        return self.__totalSourceFiles

    def __str__(self):
        return self.__strCommand

    @property
    def log(self):
        """
        class property can be used to override the class global
        logging setting

            bool:

            True if logging is enable False otherwise
        """
        if self.__log is not None:
            return self.__log

        return MKVCommandParser.classLog()

    @log.setter
    def log(self, value):
        """set instance log variable"""
        if isinstance(value, bool) or value is None:
            self.__log = value

    @property
    def analysis(self):
        """
        results of analysis of the command

        Returns:
            list:

            list with comments of anything found
        """

        return self.__lstAnalysis

    @property
    def baseFiles(self):
        lstTmp = []
        lstTmp = [x.fileName for x in self.oSourceFiles.sourceFiles]

        return lstTmp

    @property
    def bashCommand(self):
        return self.__bashCommand

    @property
    def command(self):
        return self.__strCommand

    @command.setter
    def command(self, value):
        if isinstance(value, str):
            self._initVars()
            self.__strCommand = value

            self._removeTitle()

            strCommand = convertToBashStyle(self.__strCommand)

            self.__bashCommand = strCommand
            self._parse()
            self.__readFiles = True

    @property
    def commandsGenerated(self):
        return not self.__readFiles

    @property
    def destinationFiles(self):
        """
        destination files

        Returns:
            list:

            list with destination files
        """

        if MKVParseKey.outputFile in self.filesInDirByKey:
            return self.filesInDirByKey[MKVParseKey.outputFile]

        return []

    @property
    def strCommands(self):
        return self.__strCommands

    @property
    def shellCommands(self):
        return self.__shellCommands

    def _parse(self):
        """
        _parse parse command line
        """

        strCommand = self.__bashCommand
        self.__lstAnalysis = []

        rg = r"^(.*?)\s--.*?--output.(.*?)\s--.*?\s'\('\s(.*?)\s'\)'.*?--track-order\s(.*)"

        reCommandEx = re.compile(rg)
        reExecutableEx = re.compile(r"^(.*?)\s--")
        reLanguageEx = re.compile(r"--ui-language (.*?) --")
        reOutputFileEx = re.compile(r".*?--output\s(.*?)\s--")

        reFilesEx = re.compile(
            (
                r"(?=--audio-tracks "
                r"|--video-tracks "
                r"|--subtitle-tracks "
                r"|--button-tracks "
                r"|--track-tags  "
                r"|--attachments "
                r"|--no-audio "
                r"|--no-video "
                r"|--no-subtitles "
                r"|--no-buttons "
                r"|--no-track-tags "
                r"|--no-chapters "
                r"|--no-global-tags "
                r"|--chapter-charset "
                r"|--chapter-language "
                r"|--language "
                r")"
                r"(.*?) '\(' (.*?) '\)'"
            )
        )
        reChaptersFileEx = re.compile(
            r"--chapter-language (.*?) --chapters (.*?) (?=--)"
        )
        self.__errorFound = False

        # To look Ok must match the 4 expected groups in the
        # command line
        # 1: mkvmerge name with fullpath
        # 2: output file
        # 3: at list one source
        # 4: track order
        if (matchCommand := reCommandEx.match(strCommand)) and (
            len(matchCommand.groups()) == 4  # pylint: disable=used-before-assignment
        ):
            self.cliTrackOrder = matchCommand.group(4)
            self.__lstAnalysis.append("chk: Command seems ok.")
            try:
                d = ast.literal_eval("{" + self.cliTrackOrder + "}")
                trackTotal = numberOfTracksInCommand(strCommand)
                s = self.cliTrackOrder.split(",")
                if trackTotal == len(s):
                    for e in s:
                        if not e.find(":") > 0:
                            self.__errorFound = True
                else:
                    self.__errorFound = True
                if self.__errorFound:
                    self.__lstAnalysis.append(
                        "err: Number of tracks {} and track order of {} don't match.".format(
                            trackTotal, len(d)
                        )
                    )
            except SyntaxError:
                self.__lstAnalysis.append("err: Command track order bad format.")
                self.__errorFound = True
        else:
            self.__lstAnalysis.append("err: Command bad format.")
            self.__errorFound = True

        if matchUILanguage := reLanguageEx.search(strCommand):
            self.language = matchUILanguage.group(1)

        if matchExecutable := reExecutableEx.match(strCommand):
            f = stripEncaseQuotes(matchExecutable.group(1))
            p = Path(f)
            try:
                test = p.is_file()
            except OSError:
                self.__lstAnalysis.append(
                    "err: mkvmerge incorrect syntax - {}.".format(str(p))
                )
                self.__errorFound = True
            else:
                if test:
                    self.mkvmerge = p
                    self.mkvpropedit = str(p.parent) + "mkvpropedit.exe"
                    self.__lstAnalysis.append("chk: mkvmerge ok - {}.".format(str(p)))
                else:
                    self.__lstAnalysis.append(
                        "err: mkvmerge not found - {}.".format(str(p))
                    )
                    self.__errorFound = True
        else:
            self.__lstAnalysis.append("err: mkvmerge not found.")
            self.__errorFound = True

        if matchOutputFile := reOutputFileEx.match(strCommand):
            self.cliOutputFile = None
            self.cliOutputFileMatchString = None
            f = stripEncaseQuotes(matchOutputFile.group(1))
            f = f.replace(r"'\''", "'")
            p = Path(f)
            try:
                test = Path(p.parent).is_dir()
            except OSError:
                self.__errorFound = True
                self.__lstAnalysis.append(
                    "err: Destination directory incorrect syntax - {}.".format(
                        str(p.parent)
                    )
                )
            else:
                if test:
                    self.cliOutputFile = p
                    self.cliOutputFileMatchString = matchOutputFile.group(1)
                    self.__lstAnalysis.append(
                        "chk: Destination directory ok = {}.".format(str(p.parent))
                    )
                else:
                    self.__errorFound = True
                    self.__lstAnalysis.append(
                        "err: Destination directory not found - {}.".format(
                            str(p.parent)
                        )
                    )
        else:
            self.__errorFound = True
            self.__lstAnalysis.append("err: No output file found in command.")

        if matchFiles := reFilesEx.finditer(strCommand):
            for index, match in enumerate(matchFiles):
                oFile = SourceFile(match.group(0))
                if oFile:
                    self.oSourceFiles.append(oFile)
                    if self.__totalSourceFiles is None:
                        self.__totalSourceFiles = len(oFile.filesInDir)
                    self.__lstAnalysis.append(
                        "chk: Source directory ok - {}.".format(
                            str(oFile.fileName.parent)
                        )
                    )
                    if index == 0:
                        self.filesInDirByKey[MKVParseKey.outputFile] = []
                        self.dirsByKey[MKVParseKey.outputFile] = self.cliOutputFile.parent
                        for f in oFile.filesInDir:
                            #if self.dirsByKey[MKVParseKey.outputFile] == "":
                            #    self.dirsByKey[
                            #        MKVParseKey.outputFile
                            #    ] = self.cliOutputFile.parent

                            of = self.cliOutputFile.parent.joinpath(f.stem + ".mkv")
                            of = resolveOverwrite(of)
                            self.filesInDirByKey[MKVParseKey.outputFile].append(of)
                    key = "<SOURCE{}>".format(str(index))
                    self.filesInDirByKey[key] = oFile.filesInDir
                    self.dirsByKey[key] = oFile.fileName.parent
                    if len(oFile.filesInDir) != self.__totalSourceFiles:
                        self.__errorFound = True
                        self.__lstAnalysis.append(
                            "err: Error source files TOTAL mismatched." + match.group(0)
                        )
                else:
                    self.__errorFound = True
                    self.__lstAnalysis.append(
                        "err: Error reading source files." + match.group(2)
                    )
        else:
            self.__errorFound = True
            self.__lstAnalysis.append("err: No source file found in command.")

        #
        # Optional
        #
        if matchChaptersFile := reChaptersFileEx.search(strCommand):
            self.chaptersLanguage = matchChaptersFile.group(1)
            self.cliChaptersFile = None
            self.cliChaptersFileMatchString = None
            f = unQuote(matchChaptersFile.group(2))
            p = Path(f)
            try:
                test = p.is_file()
            except OSError:
                self.__lstAnalysis.append(
                    "err: Chapters file incorrect syntax - {}.".format(str(p))
                )
                self.__errorFound = True
            else:
                if test:
                    self.cliChaptersFile = p
                    self.cliChaptersFileMatchString = matchChaptersFile.group(2)
                    self.__lstAnalysis.append(
                        "chk: Chapters file ok - {}.".format(str(p.parent))
                    )
                else:
                    self.__lstAnalysis.append(
                        "chk: Chapters file not found - {}.".format(str(p.parent))
                    )
                    self.__errorFound = True

        if self.__errorFound:
            self.__readFiles = True

    def _readDirs(self):
        """
        _readDirs read files in directories
        """

        reAttachmentsEx = re.compile(
            (
                r"--attachment-name (.*?) --attachment-mime-type "
                r"(.*?) --attach-file (.*?)(?= --)"
            )
        )

        if reAttachmentsEx.finditer(self.__bashCommand):
            self.oAttachments.strCommand = self.__bashCommand
            if self.oAttachments.isAttachmentsDirByEpisode:
                for oDir in self.oAttachments.attachmentsDirs:
                    self.__lstAnalysis.append(
                        "chk: Attachments directory ok - {}.".format(str(oDir))
                    )
            else:
                for oDir in self.oAttachments.cmdLineAttachmentsDirs:
                    self.__lstAnalysis.append(
                        "chk: Attachments directory ok - {}.".format(str(oDir))
                    )

        for f in self.oSourceFiles.sourceFiles[0].filesInDir:
            mediaInfo = MediaFileInfo(str(f))
            if mediaInfo:
                self.titles.append(mediaInfo.title)
            else:
                self.titles.append("")

        if self.cliChaptersFile:
            d = self.cliChaptersFile.parent
            fid = [x for x in d.glob("*" + self.cliChaptersFile.suffix) if x.is_file()]
            self.chaptersFiles.extend(fid)

        if self.log:
            for line in self.__lstAnalysis:
                if line.find("chk:") >= 0:
                    MODULELOG.debug("MCM0001: %s", line)
                elif line.find("err:") >= 0:
                    MODULELOG.error("MCM0002: %s", line)

    def _filesInDirByKey(self):

        if self.oAttachments.cmdLineAttachments:
            self.filesInDirByKey[
                MKVParseKey.attachmentFiles
            ] = self.oAttachments.attachmentsStr
        self.filesInDirByKey[MKVParseKey.title] = self.titles
        if self.chaptersFiles:
            self.filesInDirByKey[MKVParseKey.chaptersFile] = self.chaptersFiles

    def _template(self):

        cmdTemplate = self.__bashCommand

        reExecutableEx = re.compile(r"^(.*?)\s--")
        matchExecutable = reExecutableEx.match(cmdTemplate)
        m = matchExecutable.group(1)
        f = stripEncaseQuotes(m)
        e = shlex.quote(f)

        ##
        # BUG 1
        # Reported by zFerry98
        #
        # When running in Windows there is no space in the mkvmerge executable path
        # \ is use as escape
        # Command: ['C:binmkvtoolnixmkvmerge.exe', ...
        #
        # Solution:
        #   Force quotes for mkvmerge executable
        #
        #   Command: ['C:\\bin\\mkvtoolnix\\mkvmerge.exe'
        ##
        if platform.system() == "Windows":
            if e[0:1] != "'":
                e = "'" + f + "'"
        ##

        cmdTemplate = self.__bashCommand
        cmdTemplate = cmdTemplate.replace(m, e, 1)
        cmdTemplate = cmdTemplate.replace(
            self.cliOutputFileMatchString, MKVParseKey.outputFile, 1
        )
        for index, sf in enumerate(self.oSourceFiles.sourceFiles):
            key = "<SOURCE{}>".format(str(index))
            cmdTemplate = cmdTemplate.replace(sf.matchString, key, 1)
        if self.oAttachments.cmdLineAttachments:
            cmdTemplate = cmdTemplate.replace(
                self.oAttachments.attachmentsMatchString, MKVParseKey.attachmentFiles, 1
            )
        ##
        # Bug #3
        #
        # It was not preserving the episode title
        #
        # Remove title before parsing and added the <TITLE> key to the template
        # If there is no title read --title '' will be used.
        # working with \ ' " backslash, single and double quotes in same title
        ##

        # Add title to template
        cmdTemplate = cmdTemplate.replace(
            "--track-order", "--title " + MKVParseKey.title + " --track-order", 1
        )

        if self.cliChaptersFile:
            cmdTemplate = cmdTemplate.replace(
                self.cliChaptersFileMatchString, MKVParseKey.chaptersFile, 1
            )
        self.commandTemplate = cmdTemplate

    def _removeTitle(self):
        """
        _removeTitle remove --title option from command
        """

        ##
        # Bug #3
        #
        # It was not preserving the episode title
        #
        # Remove title before parsing and added the <TITLE> key to the template
        # If there is no title read --title '' will be used.
        # working with \ ' " backslash, single and double quotes in same title
        ##

        reTitleEx = re.compile(r".*?--title\s(.*?)\s--")

        if matchTitle := reTitleEx.match(self.__strCommand):
            self.cliTitleMatchString = matchTitle.group(1)

            self.__strCommand = self.__strCommand.replace(
                "--title " + self.cliTitleMatchString + " ", "", 1
            )

    def generateCommands(self):
        """
        generateCommands genrate and store all command lines needed
        """

        if not self.__errorFound:
            if self.__readFiles:
                self._readDirs()
                self._template()
                self._filesInDirByKey()
                self.__readFiles = False
            cmdTemplate = self.commandTemplate
            totalCommands = len(self.filesInDirByKey[MKVParseKey.outputFile])
            for i in range(totalCommands):
                keyDictionary = {}
                for key, sourceFiles in self.filesInDirByKey.items():
                    if key != MKVParseKey.attachmentFiles:
                        keyDictionary[key] = shlex.quote(str(sourceFiles[i]))
                    else:
                        keyDictionary[key] = sourceFiles[i]
                xLate = XLate(keyDictionary)  # instantiate regex dictionary translator
                strCommand = xLate.xLate(cmdTemplate)
                shellCommand = shlex.split(
                    strCommand
                )  # save command as shlex.split to submit to Pipe
                self.__strCommands.append(strCommand)
                self.__shellCommands.append(shellCommand)

    def renameOutputFiles(self, newNames):

        if len(newNames) == self.__totalSourceFiles:
            self.filesInDirByKey[MKVParseKey.outputFile] = list(newNames)
            if not self.__readFiles:
                self.generateCommands()


class MKVParseKey:

    attachmentFiles = "<ATTACHMENTS>"
    chaptersFile = "<CHAPTERS>"
    outputFile = "<OUTPUTFILE>"
    title = "<TITLE>"



"""
subclass of QAction to save text, shortcut and tooltip
this information is used for internationalization
"""

from PySide2.QtWidgets import QAction


class QActionWidget(QAction):
    """
    QActionWidget subclass of QAction save original shortcut and tooltip for locale application

    Args:
        shortcut (str, optional): original shortcut string representation. Defaults to None.
        tooltip (str, optional): original tooltip. Defaults to None.
    """

    #def __init__(self, *args, shortcut=None, tooltip=None, textPrefix=None, textSuffix=None, **kwargs):
    def __init__(self, *args, **kwargs):

        textPrefix = kwargs.pop("textPrefix", None)
        textSuffix = kwargs.pop("textSuffix", None)
        shortcut = kwargs.pop("shortcut", None)
        tooltip = kwargs.pop("tooltip", None)

        super().__init__(*args, **kwargs)

        for p in args:
            if isinstance(p, str):
                self.originaltext = p

        self.shortcut = shortcut
        self.tooltip = tooltip
        self.textPrefix = ""
        self.textSuffix = ""

        if shortcut is not None:
            self.setShortcut(shortcut)

        if tooltip is not None:
            self.setToolTip(tooltip)

        if textPrefix is not None:
            self.textPrefix = textPrefix

        if textSuffix is not None:
            self.textSuffix = textSuffix

    def setShortcut(self, shortcut, *args, **kwargs):

        if self.shortcut is None:
            self.shortcut = shortcut

        super().setShortcut(shortcut, *args, **kwargs)

    def setToolTip(self, tooltip, *args, **kwargs):

        if self.tooltip is None:
            self.tooltip = tooltip

        super().setToolTip(tooltip, *args, **kwargs)



"""
 QFrame Lines for separators
"""

from PySide2.QtWidgets import QFrame

class HorizontalLine(QFrame):
    def __init__(self, parent=None, width=1):
        super().__init__(parent)

        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Raised)
        self.setLineWidth(width)

class VerticalLine(QFrame):
    def __init__(self, parent=None, width=1):
        super().__init__(parent)

        self.setFrameShape(QFrame.VLine)
        self.setFrameShadow(QFrame.Raised)
        self.setLineWidth(width)



"""
Simple symmetric encryption based on rc4
"""

import base64

def _rc4Crypt(data, key):
    """rc4 encryption"""

    S = list(range(256))
    j = 0
    out = []

    #KSA Phase
    for i in range(256):
        j = (j + S[i] + ord( key[i % len(key)] )) % 256
        S[i] , S[j] = S[j] , S[i]

    #PRGA Phase
    i = j = 0
    for char in data:
        i = ( i + 1 ) % 256
        j = ( j + S[i] ) % 256
        S[i] , S[j] = S[j] , S[i]
        out.append(chr(ord(char) ^ S[(S[i] + S[j]) % 256]))

    return ''.join(out)

# function that encrypts data with RC4 and encodes it in base64 as default
# for other types of data encoding use a different encode parameter
# Use None for no encoding
def encrypt(data, key, encode=base64.b64encode):
    """encrypt"""

    data = _rc4Crypt(data, key)

    if encode:
        data = encode(data.encode())

    return data

#function that decrypts data with RC4 and decodes it in base64 as default
# for other types of data encoding use a different decode parameter
# Use None for no decoding
def decrypt(data, key, decode=base64.b64decode):
    """decrypt"""

    if decode:
        data = decode(data).decode()

    return _rc4Crypt(data, key)



"""
 convenience classes
"""

import re

from pathlib import Path

from natsort import natsorted, ns

from ..mkvutils import unQuote


class SourceFile:
    """
     Tracks and file names
    """

    def __init__(self, fullMatchString=None):

        self.__fullMatchString = None
        self.__errorFound = False
        self.options = None
        self.tracks = []
        self.trackID = {}
        self.filesInDir = []
        self.fileName = None
        self.matchString = None

        # for iterator
        self.__index = 0

        self.fullMatchString = fullMatchString

    def __contains__(self, item):
        return item in self.filesInDir

    def __getitem__(self, index):
        return self.filesInDir[index]

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.filesInDir)

    def __next__(self):
        if self.__index >= len(self.filesInDir):
            self.__index = 0
            raise StopIteration
        else:
            self.__index += 1
            return self.__getitem__(self.__index - 1)

    def _parse(self):

        reTrackEx = re.compile(r"(?=--language )(.*?) (?=--language|'\(')")
        reOptionsEx = re.compile(
            (
                r"(?=--audio-tracks "
                r"|--video-tracks "
                r"|--subtitle-tracks "
                r"|--button-tracks "
                r"|--track-tags  "
                r"|--attachments "
                r"|--no-audio "
                r"|--no-video "
                r"|--no-subtitles "
                r"|--no-buttons "
                r"|--no-track-tags "
                r"|--no-chapters "
                r"|--no-global-tags "
                r"|--chapter-charset "
                r"|--chapter-language "
                r"|--language "
                r")(.*?) (?='\(')"
            )
        )
        reTrackIDEx = re.compile(r"--language (\d+):(.*?)\W*")
        reTrackNameEx = re.compile(r"--track-name (.*)( --|'|)")
        reSourcesEx = re.compile(r"'\('\s(.*?)\s'\)'")
        self.tracks = []
        self.fileName = None

        if self.__fullMatchString:
            if match := reTrackEx.finditer(self.__fullMatchString):
                for i, m in enumerate(match):
                    self.tracks.append(m.group(1))
                    if trackID := reTrackIDEx.search(m.group(1)):
                        self.trackID[i] = [trackID.group(1), trackID.group(2)]
                        if trackName := reTrackNameEx.search(m.group(1)):
                            self.trackID[i].append(trackName.group(1))
                        else:
                            self.trackID[i].append("")
            if match := reOptionsEx.search(self.__fullMatchString):
                self.options = match.group(1)
            else:
                self.__errorFound = True
            if match := reSourcesEx.search(self.__fullMatchString):
                f = unQuote(match.group(1))
                p = Path(f)

                try:
                    test = p.is_file()
                except OSError:
                    self.__errorFound = True
                else:
                    if test:
                        self.fileName = p
                        self.matchString = match.group(0)
                        d = p.parent
                        fid = [x for x in d.glob("*" + p.suffix) if x.is_file()]
                        # fid.sort(key=strPath)
                        fid = natsorted(fid, alg=ns.PATH)
                        self.filesInDir.extend(fid)
                    else:
                        self.__errorFound = True
            else:
                self.__errorFound = True

    def __bool__(self):
        return not self.__errorFound

    @property
    def baseFile(self):
        return self.fileName

    @property
    def directory(self):
        return self.fileName.parent

    @property
    def fullMatchString(self):
        return self.__fullMatchString

    @fullMatchString.setter
    def fullMatchString(self, value):
        if isinstance(value, str):
            self.__fullMatchString = value
            self._parse()


class SourceFiles:
    """
     One sequence of SourceFile class elements
    """

    def __init__(self):

        self.__sourceFiles = []

        # for iterator
        self.__index = 0

    def __bool__(self):
        return bool(self.__sourceFiles)

    def __contains__(self, item):
        return item in self.__sourceFiles

    def __getitem__(self, index):

        if isinstance(index, (int, slice)):
            tmp = []
            for e in self.__sourceFiles:
                tmp.append(e[index])
            return tmp

        if isinstance(index, tuple):
            return self.__sourceFiles[index[0]]

        raise IndexError

    def __iter__(self):
        return self

    def __len__(self):
        if self.__sourceFiles:
            return len(self.__sourceFiles[0])
        return len(self.__sourceFiles)

    def __next__(self):
        if self.__index >= len(self.__sourceFiles[0]):
            self.__index = 0
            raise StopIteration
        else:
            self.__index += 1
            return self.__getitem__(self.__index - 1)

    def append(self, value):

        if isinstance(value, SourceFile):
            self.__sourceFiles.append(value)
        else:
            raise TypeError

    @property
    def sourceFiles(self):
        return self.__sourceFiles



"""
ComboBox sub-class of QComboBox permits removal of highlighted item in the popup
    list with Del key
"""


from PySide2.QtCore import Qt, Signal, Slot, QObject, QEvent
from PySide2.QtWidgets import QApplication, QComboBox


class QComboLineEdit(QComboBox):
    """
    ComboBox removes highlighted item in popup with Del key

    Args:
        parent (QWidget): parent widget
        popup (bool): confirm removal if True. No confirmation otherwise.
    """

    itemsChangeSignal = Signal()

    def __init__(self, parent=None, popup=False):
        super().__init__()

        self.__tab = None
        self.__popup = None
        self.__doPopup = popup
        self.__highlighted = None
        self.parent = parent
        self.view().installEventFilter(self)
        self.highlighted.connect(self.saveHighlighted)
        self.setEditable(True)
        self.lineEdit().setClearButtonEnabled(True)


    def isPopup(self):
        return self.__popup

    def showPopup(self):

        self.__popup = True
        super().showPopup()

    def hidePopup(self):

        self.__popup = False
        super().hidePopup()

    def eventFilter(self, obj, event):

        if obj == self.view():
            if event.type() == QEvent.KeyPress:
                if event.key() == Qt.Key_Delete:
                    if self.__highlighted is not None:
                        index = self.__highlighted
                        self.removeItem(index)
                        self.itemsChangeSignal.emit()
                        return True
            return False

        return QObject.eventFilter(self, obj, event)

    def keyPressEvent(self, event):

        if event.key() in [Qt.Key_Enter, Qt.Key_Return]:
            modifiers = QApplication.keyboardModifiers()
            if modifiers == Qt.ControlModifier:
                QComboBox.keyPressEvent(self, event)
                self.itemsChangeSignal.emit()
        else:
            QComboBox.keyPressEvent(self, event)

    @Slot(int)
    def saveHighlighted(self, index):

        self.__highlighted = index



"""
OutputTextWidget:

Output widget form just to output text in color

"""
# OTW0004 Next log ID

import logging


from PySide2.QtCore import Qt, Signal, Slot
from PySide2.QtGui import QTextCursor
from PySide2.QtWidgets import QTextEdit


from .insertTextHelpers import checkColor, LineOutput


MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class QOutputTextWidget(QTextEdit):
    """Output for running queue"""

    # log state
    __log = False
    clearSignal = Signal()
    insertTextSignal = Signal(str, dict)
    setCurrentIndexSignal = Signal()
    isDarkMode = False

    def __init__(self, parent=None, log=None, **kwargs):
        super().__init__(parent=parent, **kwargs)

        self.parent = parent
        self.__log = None
        self.__tab = None
        self.__tabWidget = None
        self.log = log

        self.insertTextSignal.connect(self.insertText)
        self.clearSignal.connect(super().clear)

    @classmethod
    def classLog(cls, setLogging=None):
        """
        get/set logging at class level
        every class instance will log
        unless overwritten

        Args:
            setLogging (bool):
                - True class will log
                - False turn off logging
                - None returns current Value

        Returns:
            bool:

            returns the current value set
        """

        if setLogging is not None:
            if isinstance(setLogging, bool):
                cls.__log = setLogging

        return cls.__log

    @property
    def log(self):
        """
        class property can be used to override the class global
        logging setting

        Returns:
            bool:

            True if logging is enable False otherwise
        """
        if self.__log is not None:
            return self.__log

        return QOutputTextWidget.classLog()

    @log.setter
    def log(self, value):
        """set instance log variable"""
        if isinstance(value, bool) or value is None:
            self.__log = value

    def connectToInsertText(self, objSignal):
        """Connect to signal"""

        objSignal.connect(self.insertText)

    @Slot(str, dict)
    def insertText(self, strText, kwargs):
        """
        insertText - Insert text in output window.
        Cannot use standard keyword arguments on
        emit calls using a dictionary argument instead

        Args:
            strText (str): text to insert on output window
            kwargs (dict): additional arguments in dictionary
                           used like **kwargs
        """

        strTmp = ""

        color = kwargs.pop(LineOutput.Color, None)
        replaceLine = kwargs.pop(LineOutput.ReplaceLine, False)
        appendLine = kwargs.pop(LineOutput.AppendLine, False)
        appendEnd = kwargs.pop(LineOutput.AppendEnd, False)
        logOnly = kwargs.pop(LineOutput.LogOnly, False)
        overrideLog = kwargs.pop("log", None)

        # still no restore to default the ideal configuration
        # search will continue considering abandoning color

        if not logOnly:
            color = checkColor(color, QOutputTextWidget.isDarkMode)

            if replaceLine:
                self.moveCursor(QTextCursor.StartOfLine, QTextCursor.KeepAnchor)
            elif appendEnd:
                self.moveCursor(QTextCursor.End)

            if color is not None:
                self.setTextColor(color)

            if replaceLine:
                self.insertPlainText(strText)
            elif appendLine:
                self.append(strText)
            elif appendEnd:
                self.append(strText)
            else:
                self.insertPlainText(strText)

            self.ensureCursorVisible()

        log = self.log

        if overrideLog is not None:
            log = overrideLog

        if log or logOnly:
            strTmp = strTmp + strText
            strTmp = strTmp.replace("\n", " ")

            if strTmp != "" and strTmp.find(u"Progress:") != 0:
                if strTmp.find(u"Warning") == 0:
                    MODULELOG.warning("OTW0001: %s", strTmp)
                elif strTmp.find(u"Error") == 0 or color == Qt.red:
                    MODULELOG.error("OTW0002: %s", strTmp)
                else:
                    if strTmp.strip():
                        MODULELOG.debug("OTW0003: %s", strTmp)



"""
rotating logging handler
it rotates at initialization
this means by number of executions
"""


import codecs
import logging.handlers
import re
import sys
from pathlib import Path


class LogRotateFileHandler(logging.handlers.RotatingFileHandler):

    """
    logging handler that rotate files at initialization time
    this will create a new file each time is run against
    the same log file.

    Args:
        fileName (Path|str): file where to save the log

        **kwargs: variable number of key value parameters
            that are passed to the super class on
            initialization
    """

    def __init__(self, fileName, **kwargs):

        # Python 3.5 open not compatible with pathlib
        if sys.version_info[:2] == (3, 5):
            f = str(fileName)
        else:
            f = fileName

        super(LogRotateFileHandler, self).__init__(f, **kwargs)

        self.doRollover()


class LogRotateFileHandlerOriginal(logging.Handler):
    """
    Custom logging handler that rotate files at the start

    :param logFile: file where to save logs
    :type logFile: str
    :param backupCount: number of files to save
    :type backupCount: int
    """

    def __init__(self, logFile, backupCount=0):
        super(LogRotateFileHandlerOriginal, self).__init__()

        self.logFile = logFile
        self.backupCount = backupCount
        self._rollover()
        self.logFilePointer = None

    def _rollover(self):
        """Rollover log files"""
        regEx = re.compile(r".*\.log\.(\d+)")

        bRotate = False
        logFile = Path(self.logFile)

        if logFile.is_file():
            bRotate = True
        # else:
        #    logFile.touch(exist_ok=True)

        if bRotate:
            filesDir = Path(logFile.parent)
            logFilesInDir = filesDir.glob("*.log.*")

            dFiles = {}
            maxIndexFile = 0

            for lFile in logFilesInDir:
                m = regEx.search(str(lFile))
                if m:
                    n = int(m.group(1))
                    dFiles[n] = lFile

            maxIndexFile = self.backupCount

            for i in range(1, self.backupCount, 1):
                if i not in dFiles:
                    maxIndexFile = i
                    break

            for n in range(maxIndexFile - 1, 0, -1):
                if n in dFiles:
                    rollFile = str(logFile) + "." + str(n + 1)
                    dFiles[n].replace(rollFile)

            rollFile = str(logFile) + ".1"
            logFile.replace(rollFile)
            # logFile.touch(exist_ok=True)

    def emit(self, record):
        """
        Write record entry to log file
        """

        # Python 3.5 open not compatible with pathlib
        if sys.version_info[:2] == (3, 5):
            f = str(self.logFile)
        else:
            f = self.logFile

        with codecs.open(f, "a", encoding="utf-8") as file:
            logEntry = self.format(record)
            file.write(logEntry + "\n")



"""
class QTColor defines the recognize SVG color keywords

colors https://en.wikipedia.org/wiki/Web_colors

"""

from PySide2.QtGui import QColor

class SvgColor:
    """
    Standard SVG color labels
    """

    aliceblue = QColor(240, 248, 255)
    antiquewhite = QColor(250, 235, 215)
    aqua = QColor(0, 255, 255)
    aquamarine = QColor(127, 255, 212)
    azure = QColor(240, 255, 255)
    beige = QColor(245, 245, 220)
    bisque = QColor(255, 228, 196)
    black = QColor(0, 0, 0)
    blanchedalmond = QColor(255, 235, 205)
    blue = QColor(0, 0, 255)
    blueviolet = QColor(138, 43, 226)
    brown = QColor(165, 42, 42)
    burlywood = QColor(222, 184, 135)
    cadetblue = QColor(95, 158, 160)
    chartreuse = QColor(127, 255, 0)
    chocolate = QColor(210, 105, 30)
    coral = QColor(255, 127, 80)
    cornflowerblue = QColor(100, 149, 237)
    cornsilk = QColor(255, 248, 220)
    crimson = QColor(220, 20, 60)
    cyan = QColor(0, 255, 255)
    darkblue = QColor(0, 0, 139)
    darkcyan = QColor(0, 139, 139)
    darkgoldenrod = QColor(184, 134, 11)
    darkgray = QColor(169, 169, 169)
    darkgreen = QColor(0, 100, 0)
    darkgrey = QColor(169, 169, 169)
    darkkhaki = QColor(189, 183, 107)
    darkmagenta = QColor(139, 0, 139)
    darkolivegreen = QColor(85, 107, 47)
    darkorange = QColor(255, 140, 0)
    darkorchid = QColor(153, 50, 204)
    darkred = QColor(139, 0, 0)
    darksalmon = QColor(233, 150, 122)
    darkseagreen = QColor(143, 188, 143)
    darkslateblue = QColor(72, 61, 139)
    darkslategray = QColor(47, 79, 79)
    darkslategrey = QColor(47, 79, 79)
    darkturquoise = QColor(0, 206, 209)
    darkviolet = QColor(148, 0, 211)
    deeppink = QColor(255, 20, 147)
    deepskyblue = QColor(0, 191, 255)
    dimgray = QColor(105, 105, 105)
    dimgrey = QColor(105, 105, 105)
    dodgerblue = QColor(30, 144, 255)
    firebrick = QColor(178, 34, 34)
    floralwhite = QColor(255, 250, 240)
    forestgreen = QColor(34, 139, 34)
    fuchsia = QColor(255, 0, 255)
    gainsboro = QColor(220, 220, 220)
    ghostwhite = QColor(248, 248, 255)
    gold = QColor(255, 215, 0)
    goldenrod = QColor(218, 165, 32)
    gray = QColor(128, 128, 128)
    grey = QColor(128, 128, 128)
    green = QColor(0, 128, 0)
    greenyellow = QColor(173, 255, 47)
    honeydew = QColor(240, 255, 240)
    hotpink = QColor(255, 105, 180)
    indianred = QColor(205, 92, 92)
    indigo = QColor(75, 0, 130)
    ivory = QColor(255, 255, 240)
    khaki = QColor(240, 230, 140)
    lavender = QColor(230, 230, 250)
    lavenderblush = QColor(255, 240, 245)
    lawngreen = QColor(124, 252, 0)
    lemonchiffon = QColor(255, 250, 205)
    lightblue = QColor(173, 216, 230)
    lightcoral = QColor(240, 128, 128)
    lightcyan = QColor(224, 255, 255)
    lightgoldenrodyellow = QColor(250, 250, 210)
    lightgray = QColor(211, 211, 211)
    lightgreen = QColor(144, 238, 144)
    lightgrey = QColor(211, 211, 211)
    lightpink = QColor(255, 182, 193)
    lightsalmon = QColor(255, 160, 122)
    lightseagreen = QColor(32, 178, 170)
    lightskyblue = QColor(135, 206, 250)
    lightslategray = QColor(119, 136, 153)
    lightslategrey = QColor(119, 136, 153)
    lightsteelblue = QColor(176, 196, 222)
    lightyellow = QColor(255, 255, 224)
    lime = QColor(0, 255, 0)
    limegreen = QColor(50, 205, 50)
    linen = QColor(250, 240, 230)
    magenta = QColor(255, 0, 255)
    maroon = QColor(128, 0, 0)
    mediumaquamarine = QColor(102, 205, 170)
    mediumblue = QColor(0, 0, 205)
    mediumorchid = QColor(186, 85, 211)
    mediumpurple = QColor(147, 112, 219)
    mediumseagreen = QColor(60, 179, 113)
    mediumslateblue = QColor(123, 104, 238)
    mediumspringgreen = QColor(0, 250, 154)
    mediumturquoise = QColor(72, 209, 204)
    mediumvioletred = QColor(199, 21, 133)
    midnightblue = QColor(25, 25, 112)
    mintcream = QColor(245, 255, 250)
    mistyrose = QColor(255, 228, 225)
    moccasin = QColor(255, 228, 181)
    navajowhite = QColor(255, 222, 173)
    navy = QColor(0, 0, 128)
    oldlace = QColor(253, 245, 230)
    olive = QColor(128, 128, 0)
    olivedrab = QColor(107, 142, 35)
    orange = QColor(255, 165, 0)
    orangered = QColor(255, 69, 0)
    orchid = QColor(218, 112, 214)
    palegoldenrod = QColor(238, 232, 170)
    palegreen = QColor(152, 251, 152)
    paleturquoise = QColor(175, 238, 238)
    palevioletred = QColor(219, 112, 147)
    papayawhip = QColor(255, 239, 213)
    peachpuff = QColor(255, 218, 185)
    peru = QColor(205, 133, 63)
    pink = QColor(255, 192, 203)
    plum = QColor(221, 160, 221)
    powderblue = QColor(176, 224, 230)
    purple = QColor(128, 0, 128)
    red = QColor(255, 0, 0)
    rosybrown = QColor(188, 143, 143)
    royalblue = QColor(65, 105, 225)
    saddlebrown = QColor(139, 69, 19)
    salmon = QColor(250, 128, 114)
    sandybrown = QColor(244, 164, 96)
    seagreen = QColor(46, 139, 87)
    seashell = QColor(255, 245, 238)
    sienna = QColor(160, 82, 45)
    silver = QColor(192, 192, 192)
    skyblue = QColor(135, 206, 235)
    slateblue = QColor(106, 90, 205)
    slategray = QColor(112, 128, 144)
    snow = QColor(255, 250, 250)
    springgreen = QColor(0, 255, 127)
    steelblue = QColor(70, 130, 180)
    tan = QColor(210, 180, 140)
    teal = QColor(0, 128, 128)
    thistle = QColor(216, 191, 216)
    tomato = QColor(255, 99, 71)
    turquoise = QColor(64, 224, 208)
    violet = QColor(238, 130, 238)
    wheat = QColor(245, 222, 179)
    white = QColor(255, 255, 255)
    whitesmoke = QColor(245, 245, 245)
    yellow = QColor(255, 255, 0)
    yellowgreen = QColor(154, 205, 50)



"""
Tabs Widget

Central widget holds:

    - ListViewWidget
    - TableViewWidget
    - TreeViewWidget

Todo: Hide and un hide at current position.  Possibly the widget itself do
      the hiding
"""
# TWG0001

import logging

from PySide2.QtCore import Signal
from PySide2.QtWidgets import QTabWidget

# from ..utils import Text

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class TabWidget(QTabWidget):
    """Main Widget"""

    # Class logging state
    __log = False

    setCurrentIndexSignal = Signal(int)
    setCurrentWidgetSignal = Signal(object)

    def __init__(self, parent=None, tabWidgets=None):
        super().__init__()

        self.parent = parent
        self.__tabWidgets = []

        self.__log = False

        self._initUI(tabWidgets)

        self.setCurrentIndexSignal.connect(super().setCurrentIndex)
        self.setCurrentWidgetSignal.connect(super().setCurrentWidget)

    def _initUI(self, tabWidgets):
        """Setup Widget Layout"""

        if tabWidgets is not None:
            self.__tabWidgets.extend(tabWidgets)

            for tw in tabWidgets:
                tabIndex = self.addTab(tw[0], tw[1])
                self.setTabToolTip(tabIndex, tw[2])

                try:
                    tw[0].tab = tabIndex
                except:  # pylint: disable=bare-except
                    pass

                try:
                    tw[0].title = tw[1]
                except:  # pylint: disable=bare-except
                    pass

                try:
                    tw[0].tabWidget = self
                except:  # pylint: disable=bare-except
                    pass

    @classmethod
    def classLog(cls, setLogging=None):
        """
        get/set logging at class level
        every class instance will log
        unless overwritten

        Args:
            setLogging (bool):
                - True class will log
                - False turn off logging
                - None returns current Value

        Returns:
            bool:
                returns the current value set
        """

        if setLogging is not None:
            if isinstance(setLogging, bool):
                cls.__log = setLogging

        return cls.__log

    @property
    def log(self):
        """
        class property can be used to override the class global
        logging setting

        Returns:
            bool:

            True if logging is enable False otherwise
        """
        if self.__log is not None:
            return self.__log

        return TabWidget.classLog()

    @log.setter
    def log(self, value):
        """set instance log variable"""
        if isinstance(value, bool) or value is None:
            self.__log = value

    def tabInserted(self, index):

        for tabIndex in range(self.count()):
            widget = self.widget(tabIndex)
            widget.tab = tabIndex

    def tabRemoved(self, index):

        for tabIndex in range(self.count()):
            widget = self.widget(tabIndex)
            widget.tab = tabIndex

    def addTabs(self, tabWidgets):
        self._initUI(tabWidgets)

    def setLanguage(self):
        """
        setLanguage set tabs labels according to locale
        """

        for i, (__, label, tooltip) in enumerate(self.__tabWidgets):
            self.setTabText(i, _(label))
            self.setTabToolTip(i, _(tooltip))



"""
subclass of QMenu to save title
used in internationalization
"""

from PySide2.QtWidgets import QMenu

class QMenuWidget(QMenu):
    """Override QMenu __init__ to save title"""

    def __init__(self, title=None, titlePrefix=None, titleSuffix=None):
        super().__init__(title)

        self.originaltitle = title
        self.titlePrefix = "" if titlePrefix is None else titlePrefix
        self.titleSuffix = "" if titleSuffix is None else titleSuffix



"""
XLate class multi regex substitution using a dictionary

Returns:
    str: text with any successful substitution
"""

import re


class XLate(dict):
    """ All-in-one multiple-string-substitution class """

    def _makeRegex(self):
        """ Build re object based on the keys of the current dictionary """
        return re.compile("|".join(map(re.escape, self.keys())))

    def __call__(self, match):
        """ Handler invoked for each regex match """
        return self[match.group(0)]

    def xLate(self, text):
        """ Translate text, returns the modified text. """
        return self._makeRegex().sub(self, text)



"""
file related functions

findFileInPath - find a file in system PATH or
    specified path

getExecutable - find a file in system PATH and
    the standard directories for applications

getFilesList - get list of files in a directory
    can be recursive
"""

from .fileutil import (
    fileQuote,
    findFileInPath,
    getFileList,
    getDirectoryList,
    getExecutable,
    stripEncaseQuotes,
)
from .classes import ConfigurationSettings, DisplayPath



"""
DualProgressBar if for graphical representation of the inner (unit) and outer (total)
loop work

Dual progress bar intended use is horizontal position vertical layout possible

align:
    Qt.Horizontal - Horizontal layout the default
    Qt.Vertical - Vertical layout

TaskBarButtonProgress show progress bar on taskbar icon

"""

import platform

from PySide2.QtCore import Qt, Signal, Slot
from PySide2.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QLabel,
    QProgressBar,
    QSizePolicy,
)

if platform.system() == "Windows":
    from PySide2.QtWinExtras import QWinTaskbarButton, QtWin
    myAppID = "VergaraSoft.MKVBatchMultiplex.mkv.2.0.0"  # arbitrary string
    QtWin.setCurrentProcessExplicitAppUserModelID(myAppID)

class DualProgressBar(QWidget):
    """
    Dual QProgressBar for unit and total progress

    param align - Set alignment Qt.Horizontal or Qt.Vertical
    type align - Qt.AlignmentFlags
    """

    valuesChangedSignal = Signal(int, int)

    def __init__(self, parent, align=Qt.Horizontal):
        super().__init__(parent)

        self.parent = parent
        self.taskbarButton = TaskbarButtonProgress(parent)
        self._lbl = QLabel()
        self._intControls()
        self._initLayout(align)

        # Default usage is to use widget in toolbar fix size
        # at extreme right
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

    def _intControls(self):

        self.pbBarUnit = QProgressBar()
        self.pbBarUnit.setRange(0, 100)
        self.pbBarUnit.setValue(0)
        self.pbBarUnit.resize(self.pbBarUnit.sizeHint())

        self.pbBarTotal = QProgressBar()
        self.pbBarTotal.setRange(0, 100)
        self.pbBarTotal.setValue(0)
        self.pbBarTotal.resize(self.pbBarTotal.sizeHint())

        # Horizontal with horizontal elements
        self.hboxLayout = QHBoxLayout()
        # Horizontal with vertical elements
        self.vboxLayout = QHBoxLayout()

    def _hLayout(self):
        """Set horizontal layout"""

        self._lbl.setText("Progress:")

        self.pbBarUnit.setOrientation(Qt.Horizontal)
        self.pbBarTotal.setOrientation(Qt.Horizontal)

        # Assign widgets to horizontal layout
        self.hboxLayout.addWidget(self._lbl)
        self.hboxLayout.addWidget(self.pbBarUnit)
        self.hboxLayout.addWidget(self.pbBarTotal)

    def _vLayout(self):
        """Set a vertical layout"""
        self.pbBarUnit.setOrientation(Qt.Vertical)
        self.pbBarTotal.setOrientation(Qt.Vertical)

        widg1 = _VerticalBarSetup(self.pbBarUnit, "U")
        widg2 = _VerticalBarSetup(self.pbBarTotal, "T")

        self.vboxLayout.addWidget(widg1)
        self.vboxLayout.addWidget(widg2)

    def _initLayout(self, align):
        """Set layout to use"""
        if align == Qt.Vertical:
            self._vLayout()
            self.setLayout(self.vboxLayout)
        else:
            self._hLayout()
            self.setLayout(self.hboxLayout)

    @property
    def label(self):
        return self._lbl

    @label.setter
    def label(self, value):
        if isinstance(value, str):
            self._lbl.setText(value)

    @Slot(int)
    def setAlignment(self, align):
        """
        Set Alignment

        : param align: Set alignment Qt.Horizontal or Qt.Vertical
        : type align: Qt.AlignmentFlags
        """

        if align == Qt.Horizontal:
            QWidget().setLayout(self.layout())
            self._intControls()
            self._hLayout()
            self.setLayout(self.hboxLayout)

        if align == Qt.Vertical:
            QWidget().setLayout(self.layout())
            self._intControls()
            self._vLayout()
            self.setLayout(self.vboxLayout)

    @Slot(int, int)
    def setMaximum(self, maxUnit=100, maxTotal=100):
        """
        Set maximum values

        : param maxUnit: maximum value for inner loop
        : type maxUnit: int
        : param maxTotal: maximum value for outer loop
        : type maxTotal: int
        """
        self.pbBarUnit.setMaximum(maxUnit)
        self.pbBarTotal.setMaximum(maxTotal)
        if self.taskbarButton:
            self.taskbarButton.progress.setMaximum(maxTotal)

    @Slot(int, int)
    def setMinimum(self, minUnit=0, minTotal=0):
        """
        Set minimum values

        : param minUnit: minimum value for inner loop
        : type minUnit: int
        : param minTotal: minimum value for outer loop
        : type minTotal: int
        """

        self.pbBarUnit.setMinimum(minUnit)
        self.pbBarTotal.setMinimum(minTotal)
        if self.taskbarButton:
            self.taskbarButton.progress.setMinimum(minTotal)

    @Slot(int, int, int, int)
    def setRange(self, minUnit=0, maxUnit=100, minTotal=0, maxTotal=100):
        """
        Set minimum and maximum as a range


        : param minUnit: minimum value for inner loop
        : type minUnit: int
        : param maxUnit: maximum value for inner loop
        : type maxUnit: int
        : param minTotal: minimum value for outer loop
        : type minTotal: int
        : param maxTotal: maximum value for outer loop
        : type maxTotal: int
        """
        self.pbBarUnit.setRange(minUnit, maxUnit)
        self.pbBarTotal.setRange(minTotal, maxTotal)
        if self.taskbarButton:
            self.taskbarButton.progress.setRange(minTotal, maxTotal)

    @Slot(int, int)
    def setValues(self, unit, total):
        """
        Update values of progressbars

        :param unit: value for inner loop
        :type unit: int
        :param total: value for outer loop
        :type total: int
        """

        self.pbBarUnit.setValue(unit)
        self.pbBarTotal.setValue(total)
        if self.taskbarButton:
            self.taskbarButton.progress.setValue(total)

        self.valuesChangedSignal.emit(unit, total)

    @Slot()
    def reset(self):
        if self.taskbarButton:
            self.taskbarButton.progress.reset()

    @Slot()
    def initTaskbarButton(self):

        if platform.system() == "Windows":
            self.taskbarButton.initTaskbarButton()

    def setSizePolicy(self, horizontal, vertical):
        """
        Argument of QSizePolicy.Policy form Qt Documentation

        QSizePolicy.Fixed              0           The QWidget.sizeHint() is the only
                                                   acceptable alternative, so the widget can
                                                   never grow or shrink (e.g. the vertical
                                                   direction of a push button).

        QSizePolicy.Minimum            GrowFlag    The sizeHint() is minimal, and sufficient.
                                                   The widget can be expanded, but there is no
                                                   advantage to it being larger (e.g. the
                                                   horizontal direction of a push button). It
                                                   cannot be smaller than the size provided by
                                                   sizeHint().

        QSizePolicy.Maximum            ShrinkFlag  The sizeHint() is a maximum. The widget can
                                                   be shrunk any amount without detriment if
                                                   other widgets need the space (e.g. a separator
                                                   line). It cannot be larger than the size
                                                   provided by sizeHint().

        QSizePolicy.Preferred          GrowFlag |ShrinkFlag
                                                   The sizeHint() is best, but the widget can be
                                                   shrunk and still be useful. The widget can be
                                                   expanded, but there is no advantage to it
                                                   being larger than sizeHint() (the default
                                                   QWidget policy).

        QSizePolicy.Expanding          GrowFlag | ShrinkFlag | ExpandFlag
                                                   The sizeHint() is a sensible size, but the widget
                                                   can be shrunk and still be useful. The widget can
                                                   make use of extra space, so it should get as much
                                                   space as possible (e.g. the horizontal direction
                                                   of a horizontal slider).
        QSizePolicy.MinimumExpanding   GrowFlag | ExpandFlag
                                                   The sizeHint() is minimal, and sufficient. The
                                                   widget can make use of extra space, so it should
                                                   get as much pace as possible (e.g. the horizontal
                                                   direction of a horizontal slider).
        QSizePolicy.Ignored            ShrinkFlag | GrowFlag | IgnoreFlag
                                                   The sizeHint() is ignored. The widget will get as
                                                   much space as possible.

        Method override is for documentation and use same name as parent also 'pass' did not
        worked as expected

        :param horizontal: horizontal size policy
        :type horizontal: QSizePolicy.Policy
        :param vertical: vertical size policy
        :type vertical: QSizePolicy.Policy
        """

        QWidget.setSizePolicy(self, horizontal, vertical)


if platform.system() == "Windows":

    class TaskbarButtonProgress(QWinTaskbarButton):
        """
        TaskbarProgress taskbar icon progress indicator
        """

        def __init__(self, parent):
            super().__init__(parent)

            self.platform = platform.system()
            self.parent = parent
            self.button = None
            self.progress = None

            if self.platform == "Windows":
                self.button = QWinTaskbarButton(parent)

        def __bool__(self):
            if (self.button is None) or (self.progress is None):
                return False
            return True

        def initTaskbarButton(self):
            """
            initTaskbarButton for late init QWinTaskbarButton
            """

            if self.platform == "Windows":
                self.button.setWindow(self.parent.windowHandle())
                self.progress = self.button.progress()
                self.progress.setRange(0, 100)
                self.progress.setVisible(True)

        @Slot(int, int)
        def setValue(self, init, total):

            self.progress.setValue(total)


else:

    def TaskbarButtonProgress(parent):
        return False


def _VerticalBarSetup(pbBar, label):
    """Vertical progress bar widget"""

    vbox = QVBoxLayout()
    widget = QWidget()

    lbl = QLabel(label)
    lbl.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)

    vbox.addWidget(pbBar)
    vbox.addWidget(lbl)

    widget.setLayout(vbox)

    return widget


class SpacerWidget(QWidget):
    """
    Utility widget to maintain widgets at same extreme during resizing
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Expanding)


if __name__ == "__main__":

    import sys
    from PySide2.QtWidgets import QApplication, QGridLayout, QMainWindow, QPushButton

    class MainWindow(QMainWindow):
        """Test the progress bars"""

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            l = QGridLayout()

            self.pb = DualProgressBar(self)

            b = QPushButton("Test")
            b.pressed.connect(self.test)
            h = QPushButton("Vertical")
            h.pressed.connect(self.vertical)
            v = QPushButton("Horizontal")
            v.pressed.connect(self.horizontal)
            l.addWidget(self.pb, 0, 0)
            l.addWidget(b, 1, 0)
            l.addWidget(h, 2, 0)
            l.addWidget(v, 3, 0)

            w = QWidget()
            w.setLayout(l)

            self.setCentralWidget(w)

            self.show()

            # this has to be defined after
            self.pb.initTaskbarButton()

        def horizontal(self):
            """Horizontal progress bar"""
            self.pb.setAlignment(Qt.Horizontal)

        def vertical(self):
            """Vertical progress bar"""
            self.pb.setAlignment(Qt.Vertical)

        def test(self):
            """Test"""
            i = 0
            j = 0
            t = 0

            self.pb.setMinimum(minTotal=0)
            self.pb.setMaximum(maxTotal=500)

            while j < 5:
                while i < 100:
                    i += 0.001
                    self.pb.setValues(i, t + i)

                t += 100
                j += 1
                i = 0

    # pylint: disable=C0103
    # variables use to construct application not constants
    app = QApplication([])
    window = MainWindow()
    sys.exit(app.exec_())



"""
Multithreading Class base on threading.Thread
"""

import logging
import multiprocessing
import traceback


MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class ProcessWorker(multiprocessing.Process):
    """
    Generic Process worker

    Args:
        function (function): function to submit to Process
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    log = False

    def __init__(self, function, *args, **kwargs):
        super(ProcessWorker, self).__init__()

        self.function = function
        self.args = args
        self.kwargs = kwargs

    def run(self):
        """
        Override run then initialise and starts the worker
        function with passed args, kwargs.
        """

        try:
            self.function(*self.args, **self.kwargs)
        except:  # pylint: disable=bare-except
            traceback.print_exc()
        return


class QueueProcessWorker(multiprocessing.Process):
    """
    Generic Queue Process worker

    Args:
        queue (Queue): Queue to process
        function (function): function to submit to Process
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    log = False

    def __init__(self, queue, function, *args, **kwargs):
        super(QueueProcessWorker, self).__init__()

        self.queue = queue
        self.function = function
        self.args = args
        self.kwargs = kwargs

    def run(self):
        """
        Override run gets next job from queue. The initialise and
        starts the worker function with passed args, kwargs and
        nextJob.
        """
        while True:
            # Get the work from the queue and expand the tuple
            nextJob = self.queue.get()
            try:
                self.function(nextJob, *self.args, **self.kwargs)
            finally:
                self.queue.task_done()



"""
miscellaneous functions/classes
"""

from .classes import XLate

from .cipher import encrypt, decrypt
from .decorators import staticVars, callCounter
from .regexutils import multipleReplace
from .strFormatTimeDelta import strFormatTimeDelta



"""
Run a function in a Thread

worker = RunInThread(function, *args, **kwargs)

# Execute
worker.run()

*args, **kwargs will be passed to the function when run


connect to finishedSignal, startSignal, resultSignal
to deal with start of execution, end of execution and
get any result from function
"""

from PySide2.QtCore import QObject, Signal

from vsutillib.process import ThreadWorker, isThreadRunning


class QRunInThread(QObject):
    """
    RunInThread - Runs a function in a Thread

    worker = RunInThread(function, *args, **kwargs)

    # Execute
    worker.run()

    *args, **kwargs will be passed to the function when run


    connect to finishedSignal, startSignal, resultSignal
    to deal with start of execution, end of execution and
    get any result from function
    Inherits from threading.Tread to handle worker thread setup, signals and wrap-up.

    Args:
        function (function): Function to submit to Thread.
        funcFinished (function): Call back function when thread finishes.
        funcResult (function): Call back function with the result of the execution.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    finishedSignal = Signal()
    startSignal = Signal()
    resultSignal = Signal(object)

    # Class logging state
    __log = False

    def __init__(self, function, *args, **kwargs):
        super().__init__()

        self.function = function
        self.args = args
        self.kwargs = kwargs

    def run(self):
        """
        run summit jobs to worker
        """

        if not isThreadRunning(self.function.__name__):
            #
            # add:
            #    funcStart
            #    funcResult
            #    funcFinished
            #
            # parameters for ThreadWorker
            #
            jobsWorker = ThreadWorker(
                self.function,
                *self.args,
                funcStart=self.start,
                funcResult=self.result,
                funcFinished=self.finished,
                **self.kwargs,
            )

            jobsWorker.name = self.function.__name__
            jobsWorker.start()

    def start(self):
        """
        start generate signal for start of run
        """
        self.startSignal.emit()

    def finished(self):
        """
        finished generate signal for finished run
        """
        self.finishedSignal.emit()

    def result(self, funcResult):
        """
        result from jobs queue process

        Args:
            funcResult (object): messages from runJobs
        """

        self.resultSignal.emit(funcResult)



"""
Regular Expresions utils
"""

import re


def multipleReplace(aDict, text):
    # Create a regular expression from all of the dictionary keys
    regex = re.compile("|".join(map(re.escape, aDict.keys())))

    # For each match, look up the corresponding value in the dictionary
    return regex.sub(lambda match: aDict[match.group(0)], text)



"""
mkvUtils:

related to mkv application functionality
"""

import glob
import logging
import os
import platform
import re
import shlex

from pathlib import Path

from vsutillib.files import findFileInPath
from vsutillib.process import RunCommand

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


def getMKVMerge():
    """
    get the name of the mkvmerge executable in the system
    search in the standard programs directories

    Returns:
        pathlib.Path:

        fully qualified mkvmerge executable
    """

    if (currentOS := platform.system()) == "Darwin":

        if lstTest := glob.glob("/Applications/MKVToolNix*"):

            f = lstTest[0] + "/Contents/MacOS/mkvmerge"

            if (mkvmerge := Path(f)).is_file():
                return mkvmerge

    elif currentOS == "Windows":

        dirs = []
        dirs.append(os.environ.get("ProgramFiles") or "")
        dirs.append(os.environ.get("ProgramFiles(x86)") or "")

        for d in dirs:
            if search := sorted(Path(d).rglob("mkvmerge.exe")):

                if (mkvmerge := Path(search[0])).is_file():
                    return mkvmerge

    elif currentOS == "Linux":

        if search := findFileInPath("mkvmerge"):

            for s in search:

                if (mkvmerge := Path(s)).is_file():
                    return mkvmerge

    return None


def getMKVMergeVersion(mkvmerge):
    """
    get mkvmerge version

    Args:
        mkvmerge (str): mkvmerge executable with full path

    Returns:
        str:

        version of mkvmerge
    """

    s = mkvmerge

    if s[0:1] != "'" and s[-1:] != "'":
        s = shlex.quote(s)

    runCmd = RunCommand(s + " --version", regexsearch=r" v(.*?) ")

    if runCmd.run():
        return runCmd.regexmatch[0]

    return None


def stripEncaseQuotes(strFile):
    """
    Strip single quote at start and end of file name
    if they are found

    Args:
        strFile (str): file name

    Returns:
        str:

        file name without start and end single quoute
    """

    # Path or str should work
    s = str(strFile)

    if (s[0:1] == "'") and (s[-1:] == "'"):
        s = s[1:-1]

    return s

def convertToBashStyle(strCommand):
    """
    Strip escape windows chars for the command line
    in the end they won't be used in a shell
    the resulting command is bash/zh like

    Args:
        strCommand (str): command generated by mkvtoolnix-gui

    Returns:
        str:

        cli command converted to bash style
    """

    strTmp = strCommand

    if strTmp.find(r'^"^(^"') >= 0:
        # This is for cmd in Windows
        strTmp = (
            strTmp.replace("'", r"'\''")
            .replace("^", "")
            .replace("/", "\\")
            .replace('"', "'")
        )

    return str(strTmp)


def numberOfTracksInCommand(strCmd):
    """
    Every track have a --language option count
    them to know the number of tracks

    Args:
        strCmd(str): command line

    Returns:
        int:

        total number of tracks
    """

    reLanguageEx = re.compile(r"\-\-language (.*?)\s")
    matchLanguage = reLanguageEx.findall(strCmd)

    return len(matchLanguage)


def resolveOverwrite(fileName, strPrefix="new-"):

    fileNameTmp = fileName

    # Check if destination file exist and add prefix if it does
    if fileNameTmp.is_file():

        strSuffix = ""
        n = 1

        while True:
            fileNameTmp = fileNameTmp.parent.joinpath(
                strPrefix + fileName.stem + strSuffix + fileName.suffix
            )

            if fileNameTmp.is_file():
                strSuffix = " ({})".format(n)
                n += 1
            else:
                break

    return fileNameTmp


def unQuote(fileName):
    """
    Remove start end quotes and escape ones
    Args:
        fileName (str): file name

    Returns:
        str:

        file name without quotes if found
    """

    f = stripEncaseQuotes(fileName)
    f = f.replace(r"'\''", "'")

    return f

def strPath(value):
    """
    Convenience function to help sort

    Arguments:
        value {various types} -- value to convert to string

    Returns:
        str -- argument received converted to string
    """
    return str(value)



"""
Verify structure of media files for inconsistencies
against the source base files use for the templates
"""
# VFS0001

import logging

from pathlib import Path

from vsutillib.media import MediaFileInfo

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class VerifyStructure:
    """
    Convenience class use by MKVCommand_ to verify structure
    of media files against base files.

    The class evaluates to True if structure if logically equal.
    That is tracks same order same type.

    .. code:: Python

        verify = VerifyStructure(lstBaseFile, lstFiles)

        if verify:
            # Ok to proceed
            ...
        else:
            raise ValueError('')

    Args:
        lstBaseFile (:obj:`list`, optional): list with the base files
            as found in command
        lstFiles (:obj:`list`, optional): list of files to generate new command
    """

    __log = False

    def __init__(
        self, lstBaseFiles=None, lstFiles=None, destinationFile=None, log=None
    ):

        self.__analysis = None
        self.__log = None
        self.__status = None

        self.log = log

        if (lstBaseFiles is not None) and (lstFiles is not None):
            self.verifyStructure(lstBaseFiles, lstFiles, destinationFile)

    def __bool__(self):
        return self.__status

    def __str__(self):

        return "".join(self.__analysis)

    @classmethod
    def classLog(cls, setLogging=None):
        """
        get/set logging at class level
        every class instance will log
        unless overwritten

        Args:
            setLogging (`bool`):

                - True class will log
                - False turn off logging
                - None returns current Value

        Returns:
            bool:

            returns the current value set
        """

        if setLogging is not None:
            if isinstance(setLogging, bool):
                cls.__log = setLogging

    @property
    def log(self):
        """
        class property can be used to override the class global
        logging setting if set to None class log will be followed

        Returns:
            bool:

            True if logging is enable False otherwise
        """
        if self.__log is not None:
            return self.__log

        return VerifyStructure.classLog()

    @log.setter
    def log(self, value):
        """set instance log variable"""
        if isinstance(value, bool) or value is None:
            self.__log = value

    @property
    def isOk(self):
        """
        status check of analysis

        Returns:
            bool:

            Returns True if successful False otherwise.
        """
        return self.__status

    @property
    def analysis(self):
        """
        results of analysis of the comparison

        Returns:
            list:

            list with comments of anything found
        """
        return self.__analysis

    def verifyStructure(self, lstBaseFiles, lstFiles, destinationFile=None):
        """
        Verify if structure of files if logically equal.


        Args:
            lstBaseFile (list): list with the base files
                as found in command
            lstFiles (list): list of files to generate new command
        """

        # msg = "Error: In structure \n\nSource:\n{}\nBase Source:\n{}\n"
        self.__analysis = []
        self.__status = True

        for strSource, strFile in zip(lstBaseFiles, lstFiles):

            try:

                objSource = MediaFileInfo(strSource, log=self.log)
                objFile = MediaFileInfo(strFile, log=self.log)

                if objSource != objFile:

                    if destinationFile is not None:
                        msg = "Error: In structure\nDestination File: {}\n\n"
                        msg = msg.format(destinationFile)
                    else:
                        msg = "Error: In structure\n\n"
                    msg = msg + "Source:\n{}\n\nBase Source:\n{}\n"
                    msg = msg.format(str(objFile), str(objSource))
                    self.__analysis.append(msg)
                    self.__status = False
                    _detailAnalysis(self.__analysis, objFile, objSource)

                    if self.log:

                        msg = "Error: In structure Source: {} Base Source: {}"
                        msg = msg.format(objFile.fileName, objSource.fileName)
                        MODULELOG.error("VFS0002: Error: %s", msg)

                        for i, line in enumerate(self.__analysis):
                            if i > 0:
                                MODULELOG.error("VFS0003: Error: %s", line.strip())

                        msg = "Structure not ok. Source: {} Base Source: {}"
                        msg = msg.format(objFile.fileName, objSource.fileName)
                        MODULELOG.debug("VFS0004: %s", msg)

                else:
                    if self.log:
                        msg = "Structure seems ok. Source: {} Base Source: {}"
                        msg = msg.format(objFile.fileName, objSource.fileName)
                        MODULELOG.debug("VFS0005: %s", msg)

            except OSError as error:

                msg = "Error: \n{}\n"
                msg = msg.format(error.strerror)
                self.__analysis.append(msg)
                self.__status = False

                if self.log:
                    msg = "Error: {}"
                    msg = msg.format(error.strerror)
                    MODULELOG.error("VFS0001: %s", msg)


def _detailAnalysis(lstAnalysis, mediaFile1, mediaFile2):

    name1 = Path(mediaFile1.fileName).name
    name2 = Path(mediaFile2.fileName).name

    if mediaFile1.codec != mediaFile2.codec:
        msg = "Codec mismatched {}: {} - {}: {}\n".format(
            name1, mediaFile1.codec, name2, mediaFile2.codec
        )
        lstAnalysis.append(msg)
    elif len(mediaFile1) != len(mediaFile2):
        msg = "Number of tracks mismatched {}: {} - {}: {}\n".format(
            name1, len(mediaFile1), name2, len(mediaFile2)
        )
        lstAnalysis.append(msg)
    elif len(mediaFile1) == len(mediaFile2):
        for a, b in zip(mediaFile1.lstMediaTracks, mediaFile2.lstMediaTracks):
            if a.streamorder != b.streamorder:
                msg = "Stream order mismatched {}: {} - {}: {}\n".format(
                    name1, a.streamorder, name2, b.streamorder
                )
                lstAnalysis.append(msg)
            elif a.track_type != b.track_type:
                msg = "Stream type mismatched {}: {} - {}: {}\n".format(
                    name1, a.track_type, name2, b.track_type
                )
                lstAnalysis.append(msg)
            elif a.language != b.language:
                msg = "Stream language mismatched {}: {}:{} - {}: {}:{}\n".format(
                    name1, a.streamorder, a.language, name2, b.streamorder, b.language
                )
                lstAnalysis.append(msg)
            elif (a.codec != b.codec) and (a.track_type != "Audio"):
                msg = "Codec mismatched {}: {} - {}: {}\n".format(
                    name1, a.codec, name2, b.codec
                )
                lstAnalysis.append(msg)
            elif a.format != b.format:
                msg = "Stream format mismatched {}: {} - {}: {}\n".format(
                    name1, a.format, name2, b.format
                )
                lstAnalysis.append(msg)



"""
PySide2 related classes and functions
"""

# classes

from .classes import (
    checkColor,
    DualProgressBar,
    HorizontalLine,
    LineOutput,
    QActionWidget,
    QComboLineEdit,
    QFileListWidget,
    QFormatLabel,
    QGroupSignal,
    QMenuWidget,
    QOutputTextWidget,
    QProgressIndicator,
    QPushButtonWidget,
    QRunInThread,
    QSignalLogHandler,
    QSystemTrayIconWidget,
    QthThread,
    QthThreadWorker,
    SvgColor,
    TabWidget,
    TabWidgetExtension,
    TaskbarButtonProgress,
    VerticalLine,
    Worker,
    WorkerSignals,
)

from .messagebox import messageBox, messageBoxYesNo

from .qtUtils import (
    centerWidget,
    pushButton,
    darkPalette,
    qtRunFunctionInThread,
)



"""
Convenience functions related to files and file system
"""

import os
import platform
import shlex

from pathlib import Path, PurePath


def findFileInPath(element, dirPath=None):
    """
    Convenience function that finds a file in the
    PATH environment variable

    Args:
        dirPath (str): search in the specified PATH

    Returns:
        list: list of :py:class:`pathlib.Path`

    I really like the :mod:`threading` module which has the
    :class:`threading.Thread` class.

    Here is a link :func:`time.time`.
    """

    filesFound = []

    if dirPath is None:
        dirPath = os.getenv("PATH")

    if isinstance(dirPath, str):
        dirs = dirPath.split(os.pathsep)
    else:
        dirs = dirPath

    for dirname in dirs:
        candidate = Path(PurePath(dirname).joinpath(element))
        if candidate.is_file():
            filesFound.append(candidate)

    return filesFound


def getFileList(
    strPath, wildcard="*.*", fullpath=False, recursive=False, strName=False
):
    """
    Get files in a directory
    strPath has to be an existing directory or file
    in case of a file the parent directory is used
    strExtFilter in the form -> .ext
    """

    doubleWildcard = False
    argWildcard = wildcard
    lstObjFileNames = []

    p = Path(strPath)

    try:
        if p.is_file():
            p = p.parent
    except OSError:
        pass

    try:
        if not p.is_dir():
            p = p.parent
    except OSError:
        wildcard = p.stem
        p = p.parent
        doubleWildcard = True  # wildcard found on strPath argument

    try:
        if not p.is_dir():
            # Wrong argument for strPath
            return []
    except OSError:
        # Wrong argument for strPath
        return []

    if (not p.is_file()) and (not p.is_dir()):
        return []

    lstFilesFilter = []

    if p.is_file():
        p = p.parent

    wc = wildcard
    awc = argWildcard

    if doubleWildcard:

        lstFiles = [x for x in p.glob(wc) if x.is_file()]

        if recursive:
            awc = "**/" + stripEncaseQuotes(argWildcard)

            for d in lstFiles:
                if d.is_dir():
                    oFiles = [x for x in d.glob(awc) if x.is_file()]
                    lstObjFileNames.extend(oFiles)
        else:
            lstObjFileNames.extend(lstFiles)

    else:
        if recursive:
            wc = "**/" + stripEncaseQuotes(wildcard)

        # print("Path = {}\nWith wildcard ={}\n".format(p, wc))

        lstObjFileNames = [x for x in p.glob(wc) if x.is_file()]

    if not fullpath:
        lstFilesFilter = [x.name for x in lstObjFileNames]
        return lstFilesFilter

    if strName:
        lstFilesFilter = [str(x) for x in lstObjFileNames]
        return lstFilesFilter

    return lstObjFileNames


def getDirectoryList(
    strPath, wildcard="*", fullpath=False, recursive=False, strName=False
):
    """
    Get files in a directory
    strPath has to be an existing directory or file
    in case of a file the parent directory is used
    strExtFilter in the form -> .ext
    """

    doubleWildcard = False
    argWildcard = wildcard
    lstObjFileNames = []

    p = Path(strPath)

    try:
        if p.is_file():
            p = p.parent
    except OSError:
        pass

    try:
        if not p.is_dir():
            p = p.parent
    except OSError:
        wildcard = p.stem
        p = p.parent
        doubleWildcard = True  # wildcard found on strPath argument

    try:
        if not p.is_dir():
            # Wrong argument for strPath
            return []
    except OSError:
        # Wrong argument for strPath
        return []

    lstFilesFilter = []

    wc = wildcard
    awc = argWildcard

    if doubleWildcard:
        dirs = [x for x in p.glob(wc) if x.is_dir()]

        if recursive:
            awc = "**/" + stripEncaseQuotes(argWildcard)
            for d in dirs:
                if d.is_dir():
                    lstObjFileNames.append(d)
                    oDirs = [x for x in d.glob(awc) if x.is_dir()]
                    lstObjFileNames.extend(oDirs)

        else:
            lstObjFileNames.extend(dirs)
    else:
        if recursive:
            wc = "**/" + stripEncaseQuotes(wildcard)

        lstObjFileNames = [x for x in p.glob(wc) if x.is_dir()]

    if not fullpath:
        lstFilesFilter = [x.name for x in lstObjFileNames]
        return lstFilesFilter

    if strName:
        lstFilesFilter = [str(x) for x in lstObjFileNames]
        return lstFilesFilter

    return lstObjFileNames


def getExecutable(search):
    """
    search for executable for macOS and
    """

    fileToSearch = search

    currentOS = platform.system()

    if currentOS == "Darwin":

        lstTest = Path("/Applications").glob("**/" + fileToSearch)

        for l in lstTest:
            if (p := Path(l)).stem == fileToSearch:
                return p

    elif currentOS == "Windows":

        if fileToSearch.find(".") < 0:
            # assume is binary executable
            fileToSearch += ".exe"

        dirs = []
        dirs.append(os.environ.get("ProgramFiles") or "")
        dirs.append(os.environ.get("ProgramFiles(x86)") or "")

        # search 64 bits
        for d in dirs:
            if search := sorted(Path(d).rglob(fileToSearch)):
                if (executable := Path(search[0])).is_file():
                    return executable

    if searchFile := findFileInPath(fileToSearch):
        for e in searchFile:
            if (executable := Path(e)).is_file():
                return executable

    return None


def stripEncaseQuotes(strFile):
    """
    Strip single quote at start and end of file name
    if they are found

    Args:
        strFile (str): file name

    Returns:
        str:

        file name without start and end single quoute
    """

    # Path or str should work
    s = str(strFile)

    if (s[0:1] == "'") and (s[-1:] == "'"):
        s = s[1:-1]

    return s


def fileQuote(oFile):

    f = shlex.quote(str(oFile))
    s = f.replace('"\'"', r"\'")

    return s



"""VS module names"""

# MKV0001

#     MKVCommandNew,

from .classes import (
    MKVAttachment,
    MKVAttachments,
    MKVCommand,
    MKVCommandParser,
    MKVParseKey,
    SourceFile,
    SourceFiles,
    VerifyMKVCommand,
    VerifyStructure,
)
from .mkvutils import (
    convertToBashStyle,
    getMKVMerge,
    getMKVMergeVersion,
    numberOfTracksInCommand,
    resolveOverwrite,
    stripEncaseQuotes,
    unQuote,
)
